# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from contextlib import contextmanager
import io
import logging
import os.path
import shutil
import sys
import tempfile

from ..errors import ValidationFailed


PY3 = sys.version_info.major == 3


logger = logging.getLogger(__name__)


def is_enabled():
    try:
        from lxml import etree  # noqa
    except ImportError:
        return False
    else:
        return True


def xslt(xsl_path, inp_path, out_path):
    ''' Transform XML with XSL

    :param xsl_path: stylesheet path
    :param inp_path: input path
    :param out_path: output path
    '''
    transform = xslt_compile(xsl_path)
    with io.open(out_path, 'wb') as f:
        return transform(inp_path, f)


def xslt_compile(xsl_path, **params):
    xslt = XSLT(xsl_path, **params)
    return xslt.transform_into_stream


class XSLT:

    def __init__(self, xsl_path, **params):
        ''' Compile XSL Transform function.
        :param xsl_path: stylesheet path
        :returns: a transform function
        '''
        from lxml import etree

        with io.open(xsl_path, 'rb') as xsl_file:
            xsl_doc = etree.parse(xsl_file)

        self.xsl_path = xsl_path
        self.etree_xslt = etree.XSLT(xsl_doc)
        self.params = dict((name, etree.XSLT.strparam(value))
                           for name, value in params.items())

    def transform(self, input, output):
        '''
        >>> T.transform('input.xml', 'output.xml')
        '''
        with io.open(input, 'rb') as inp_file:
            with io.open(output, 'wb') as out_file:
                return self._transform(inp_file, out_file)

    def transform_into_stream(self, input, output):
        '''
        >>> T.transform_into_stream('input.xml', sys.stdout)
        '''
        with io.open(input, 'rb') as inp_file:
            return self._transform(inp_file, output)

    def _transform(self, input, output):
        # input, output: binary stream

        from lxml import etree
        source = etree.parse(input)
        logger.info('_lxml.xslt(%s) start',
                    os.path.basename(self.xsl_path))
        result = self.etree_xslt(source, **self.params)
        logger.info('_lxml.xslt(%s) end',
                    os.path.basename(self.xsl_path))
        # https://lxml.de/1.3/FAQ.html#what-is-the-difference-between-str-xslt-doc-and-xslt-doc-write
        result = bytes(result)
        output.write(result)
        return dict()


def relaxng(rng_path, inp_path):
    relaxng = RelaxNG(rng_path)
    return relaxng.validate(inp_path)


def relaxng_compile(rng_path):
    ''' Compile RelaxNG file

    :param rng_path: RelaxNG path
    :returns: a validation function
    '''
    return RelaxNG(rng_path)


class RelaxNG:

    def __init__(self, rng_path):
        from lxml import etree

        with io.open(rng_path, 'rb') as rng_file:
            rng = etree.parse(rng_file)

        self.rng_path = rng_path
        self.etree_relaxng = etree.RelaxNG(rng)

    @contextmanager
    def validating_output(self, output):
        fd, name = tempfile.mkstemp()
        try:
            with os.fdopen(fd, 'wb+') as f:
                yield f
                f.seek(0)
                if not self.validate_stream(f):
                    raise ValidationFailed('RelaxNG')
                f.seek(0)
                shutil.copyfileobj(f, output)
        finally:
            try:
                os.unlink(name)
            except Exception as e:
                logger.warning('%s: can\'t unlink %s', e, name)

    def validate(self, input):
        from lxml import etree
        with io.open(input, 'rb') as f:
            doc = etree.parse(f)
        return self._validate(doc)

    def validate_stream(self, input):
        from lxml import etree
        doc = etree.parse(input)
        return self._validate(doc)

    def _validate(self, doc):
        logger.info('_lxml.relaxng(%s) start', os.path.basename(self.rng_path))
        try:
            valid = self.etree_relaxng.validate(doc)
        except Exception as e:
            logger.exception(e)
            raise
        else:
            if not valid:
                for error in self.etree_relaxng.error_log:
                    logger.error('%s', error)
            return valid
        finally:
            logger.info(
                '_lxml.relaxng(%s) end',
                os.path.basename(self.rng_path)
            )


def errlog_to_dict(error):
    return dict(message=error.message,
                filename=error.filename,
                line=error.line,
                column=error.column,
                domain=error.domain_name,
                type=error.type_name,
                level=error.level_name)



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import uno
import unohelper
from com.sun.star.io import XInputStream, XSeekable, XOutputStream


class InputStreamFromFileLike(unohelper.Base, XInputStream, XSeekable):
    ''' Implementation of XInputStream, XSeekable based on a file-like object

    Implements com.sun.star.io.XInputStream and com.sun.star.io.XSeekable

    :param f: a file-like object
    '''
    def __init__(self, f, dontclose=False):
        self.f = f
        self.dontclose = dontclose

    def readBytes(self, aData, nBytesToRead):
        data = self.f.read(nBytesToRead)
        return len(data), uno.ByteSequence(data)

    readSomeBytes = readBytes

    def skipBytes(self, nBytesToSkip):
        self.f.read(nBytesToSkip)

    def available(self):
        return 0

    def closeInput(self):
        if not self.dontclose:
            self.f.close()

    def seek(self, location):
        self.f.seek(location)

    def getPosition(self):
        pos = self.f.tell()
        return pos

    def getLength(self):
        pos = self.f.tell()
        try:
            self.f.seek(0, 2)
            length = self.f.tell()
            return length
        finally:
            self.f.seek(pos)


class OutputStreamToFileLike(unohelper.Base, XOutputStream):
    ''' Implementation of XOutputStream based on a file-like object.

    Implements com.sun.star.io.XOutputStream.

    :param f: a file-like object
    '''
    def __init__(self, f, dontclose=False):
        self.f = f
        self.dontclose = dontclose

    def writeBytes(self, bytesequence):
        self.f.write(bytesequence.value)

    def flush(self):
        self.f.flush()

    def closeOutput(self):
        if not self.dontclose:
            self.f.close()


class FileFromStream(object):
    ''' A file-like object based on XInputStream/XOuputStream/XSeekable

    :param stream: a stream object which implements
    com.sun.star.io.XInputStream, com.sun.star.io.XOutputStream or
    com.sun.star.io.XSeekable
    '''
    def __init__(self, stream):
        self.stream = stream

        if hasattr(stream, 'readBytes'):
            def read(size=None):
                if size is None:
                    data = ''
                    while True:
                        bytes = uno.ByteSequence('')
                        n_read, bytes = stream.readBytes(bytes, 4096)
                        if n_read == 0:
                            return data
                        data += bytes.value
                bytes = uno.ByteSequence('')
                n_read, bytes = stream.readBytes(bytes, size)
                return bytes.value
            self.read = read

        if hasattr(stream, 'seek'):
            self.tell = stream.getPosition

            def seek(offset, whence=0):
                if whence == 0:
                    pass
                elif whence == 1:
                    offset += stream.getPosition()
                elif whence == 2:
                    offset += stream.getLength()
                stream.seek(offset)
            self.seek = seek

        if hasattr(stream, 'writeBytes'):
            def write(s):
                stream.writeBytes(uno.ByteSequence(s))
            self.write = write

            def flush():
                stream.flush()
            self.flush = flush

    def close(self):
        if hasattr(self.stream, 'closeInput'):
            self.stream.closeInput()
        elif hasattr(self.stream, 'closeOutput'):
            self.stream.closeOutput()



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from hwp5.binmodel._shared import RecordModel
from hwp5.tagids import HWPTAG_BULLET
from hwp5.dataio import INT32
from hwp5.dataio import UINT32
from hwp5.dataio import UINT16
from hwp5.dataio import HWPUNIT16
from hwp5.dataio import WCHAR
from hwp5.dataio import Enum
from hwp5.dataio import Flags


BulletAlignEnum = Enum(LEFT=0, CENTER=1, RIGHT=2)
BulletFlags = Flags(UINT32,
                    0, 1, BulletAlignEnum, 'align',
                    3, 'auto_indent')


class Bullet(RecordModel):
    ''' 4.1.9. 글머리표 '''

    tagid = HWPTAG_BULLET

    @staticmethod
    def attributes():
        # TODO: Spec 1.2 is insufficient and incorrect
        yield BulletFlags, 'flags',
        yield HWPUNIT16, 'width',  # 너비, 단위: HWPUNIT
        yield UINT16, 'space',  # 본문과의 간격, 단위: %
        yield INT32, 'charshape_id',
        yield WCHAR, 'char'



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from hwp5.binmodel.controlchar import CHID
from hwp5.dataio import Flags
from hwp5.dataio import UINT32
from hwp5.dataio import BSTR
from hwp5.dataio import BYTE
from hwp5.binmodel.tagid71_ctrl_data import ControlData
from hwp5.binmodel.controls._shared import Control


class Field(Control):
    ''' 4.2.10.15 필드 시작 '''

    Flags = Flags(UINT32,
                  0, 'editableInReadOnly',
                  11, 14, 'visitedType',
                  15, 'modified')

    def attributes(cls):
        yield cls.Flags, 'flags',
        yield BYTE, 'extra_attr',
        yield BSTR, 'command',
        yield UINT32, 'id',
    attributes = classmethod(attributes)


class FieldUnknown(Field):
    chid = CHID.FIELD_UNK


class FieldDate(Field):
    chid = CHID.FIELD_DTE


class FieldDocDate(Field):
    chid = CHID.FIELD_DDT


class FieldPath(Field):
    chid = CHID.FIELD_PAT


class FieldBookmark(Field):
    chid = CHID.FIELD_BMK


class FieldMailMerge(Field):
    chid = CHID.FIELD_MMG


class FieldCrossRef(Field):
    chid = CHID.FIELD_XRF


class FieldFormula(Field):
    chid = CHID.FIELD_FMU


class FieldClickHere(Field):
    chid = CHID.FIELD_CLK


class FieldClickHereData(ControlData):
    parent_model_type = FieldClickHere


class FieldSummary(Field):
    chid = CHID.FIELD_SMR


class FieldUserInfo(Field):
    chid = CHID.FIELD_USR


class FieldHyperLink(Field):
    chid = CHID.FIELD_HLK

    def geturl(self):
        s = self.command.split(';')
        return s[0].replace('\\:', ':')


class FieldRevisionSign(Field):
    chid = CHID.FIELD_REVISION_SIGN


class FieldRevisionDelete(Field):
    chid = CHID.FIELD_REVISION_DELETE


class FieldRevisionAttach(Field):
    chid = CHID.FIELD_REVISION_ATTACH


class FieldRevisionClipping(Field):
    chid = CHID.FIELD_REVISION_CLIPPING


class FieldRevisionSawtooth(Field):
    chid = CHID.FIELD_REVISION_SAWTOOTH


class FieldRevisionThinking(Field):
    chid = CHID.FIELD_REVISION_THINKING


class FieldRevisionPraise(Field):
    chid = CHID.FIELD_REVISION_PRAISE


class FieldRevisionLine(Field):
    chid = CHID.FIELD_REVISION_LINE


class FieldRevisionSimpleChange(Field):
    chid = CHID.FIELD_REVISION_SIMPLECHANGE


class FieldRevisionHyperlink(Field):
    chid = CHID.FIELD_REVISION_HYPERLINK


class FieldRevisionLineAttach(Field):
    chid = CHID.FIELD_REVISION_LINEATTACH


class FieldRevisionLineLink(Field):
    chid = CHID.FIELD_REVISION_LINELINK


class FieldRevisionLineTransfer(Field):
    chid = CHID.FIELD_REVISION_LINETRANSFER


class FieldRevisionRightMove(Field):
    chid = CHID.FIELD_REVISION_RIGHTMOVE


class FieldRevisionLeftMove(Field):
    chid = CHID.FIELD_REVISION_LEFTMOVE


class FieldRevisionTransfer(Field):
    chid = CHID.FIELD_REVISION_TRANSFER


class FieldRevisionSimpleInsert(Field):
    chid = CHID.FIELD_REVISION_SIMPLEINSERT


class FieldRevisionSplit(Field):
    chid = CHID.FIELD_REVISION_SPLIT


class FieldRevisionChange(Field):
    chid = CHID.FIELD_REVISION_CHANGE


class FieldMemo(Field):
    chid = CHID.FIELD_MEMO


class FieldPrivateInfoSecurity(Field):
    chid = CHID.FIELD_PRIVATE_INFO_SECURITY



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from uuid import UUID

from .msoleprops import PropertyIdentifier
from .msoleprops import RESERVED_PROPERTIES
from .msoleprops import SUMMARY_INFORMATION_PROPERTIES


CLSID_HWP_SUMMARY_INFORMATION = UUID(
    '9fa2b660-1061-11d4-b4c6-006097c09d8c'
)

FMTID_HWP_SUMMARY_INFORMATION = CLSID_HWP_SUMMARY_INFORMATION

HWPPIDSI_DATE_STR = PropertyIdentifier(
    id=0x00000014,
    label='HWPPIDSI_DATE_STR',
)

HWPPIDSI_PARACOUNT = PropertyIdentifier(
    id=0x00000015,
    label='HWPPIDSI_PARACOUNT',
)

HWP_PROPERTIES = RESERVED_PROPERTIES + SUMMARY_INFORMATION_PROPERTIES + (
    HWPPIDSI_DATE_STR,
    HWPPIDSI_PARACOUNT,
)


class HwpSummaryInfoTextFormatter(object):

    def formatTextLines(self, hwpsummaryinfo):
        yield 'Title: {}'.format(hwpsummaryinfo.title)
        yield 'Subject: {}'.format(hwpsummaryinfo.subject)
        yield 'Author: {}'.format(hwpsummaryinfo.author)
        yield 'Keywords: {}'.format(hwpsummaryinfo.keywords)
        yield 'Comments: {}'.format(hwpsummaryinfo.comments)
        yield 'Last saved by: {}'.format(hwpsummaryinfo.lastSavedBy)
        yield 'Revision Number: {}'.format(hwpsummaryinfo.revisionNumber)
        yield 'Last Printed at: {}'.format(hwpsummaryinfo.lastPrintedTime)
        yield 'Created at: {}'.format(hwpsummaryinfo.createdTime)
        yield 'Last saved at: {}'.format(hwpsummaryinfo.lastSavedTime)
        yield 'Number of pages: {}'.format(hwpsummaryinfo.numberOfPages)
        yield 'Date: {}'.format(hwpsummaryinfo.dateString)
        yield 'Number of paragraphs: {}'.format(
            hwpsummaryinfo.numberOfParagraphs
        )



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from .binmodel import ControlChar
from .binmodel import ParaTextChunks
from .binmodel import Text
from .bintype import resolve_type_events
from .bintype import resolve_values_from_stream
from .dataio import ArrayType
from .dataio import EnumType
from .dataio import FlagsType
from .dataio import SelectiveType
from .dataio import StructType
from .dataio import X_ARRAY
from .filestructure import FileHeader
from .treeop import ENDEVENT
from .treeop import STARTEVENT
from .xmlformat import xmlevents_to_bytechunks


def xmldump_flat(hwp5file, output, xml_declaration=True):
    xmlevents = xmlevents_from_hwp5file(hwp5file)
    bytechunks = xmlevents_to_bytechunks(xmlevents)
    if xml_declaration:
        output.write('<?xml version="1.0" encoding="utf-8"?>')
    for x in bytechunks:
        output.write(x)


def xmlevents_from_hwp5file(hwp5file):
    version = '%d.%d.%d.%d' % hwp5file.fileheader.version
    hexversion = '%02x%02x%02x%02x' % hwp5file.fileheader.version
    yield STARTEVENT, ('Hwp5Doc', {
        'version': version,
        'hexversion': hexversion
    })

    with hwp5file.fileheader.open() as f:
        yield STARTEVENT, ('FileHeader', {
        })
        resolve_values = resolve_values_from_stream(f)
        model_events = resolve_type_events(FileHeader, {}, resolve_values)
        for x in xmlevents_from_modelevents(model_events):
            yield x
        yield ENDEVENT, 'FileHeader'

    yield STARTEVENT, ('DocInfo', {
    })
    model_events = hwp5file.docinfo.parse_model_events()
    for x in xmlevents_from_modelevents(model_events):
        yield x
    yield ENDEVENT, 'DocInfo'

    for section_name in hwp5file.bodytext:
        section = hwp5file.bodytext[section_name]
        yield STARTEVENT, ('Section', {
            'name': section_name
        })
        model_events = section.parse_model_events()
        for x in xmlevents_from_modelevents(model_events):
            yield x
        yield ENDEVENT, 'Section'

    yield ENDEVENT, 'Hwp5Doc'


def expand_item_value(ev, data):
    if ev is None and data['type'] is ParaTextChunks:
        yield STARTEVENT, data
        for (start, end), item in data['value']:
            if isinstance(item, unicode):
                yield None, {
                    'bin_offset': data['bin_offset'] + start * 2,
                    'type': Text,
                    'value': item,
                }
            else:
                x = {
                    'bin_offset': data['bin_offset'] + start * 2,
                    'type': ControlChar,
                    'value': item,
                }
                yield None, x
        yield ENDEVENT, data
    else:
        yield ev, data


def expand_item_values(model_events):
    for ev, item in model_events:
        for x in expand_item_value(ev, item):
            yield x


def xmlevents_from_modelevents(model_events):  # noqa
    expanded_events = expand_item_values(model_events)
    for ev, data in expanded_events:
        record = data.get('record')
        if record:
            if ev is STARTEVENT:
                yield ev, ('Record', {
                    'tagname': record['tagname'],
                    'tagid': unicode(record['tagid']),
                    'seqno': unicode(record['seqno']),
                    'level': unicode(record['level']),
                    'size': unicode(record['size']),
                })
            elif ev is ENDEVENT:
                yield ev, 'Record'
            else:
                assert False
        else:
            datatype = data['type']
            typename = datatype.__name__

            if ev in (STARTEVENT, ENDEVENT):
                if isinstance(datatype, (ArrayType, X_ARRAY)):
                    elem = 'array'
                    atrs = {
                    }
                elif isinstance(datatype, (StructType, SelectiveType)):
                    elem = 'struct'
                    atrs = {
                        'type': typename
                    }
                else:
                    raise Exception(datatype.__name__)

                if 'name' in data:
                    atrs['name'] = data['name']
                if ev is STARTEVENT:
                    yield ev, (elem, atrs)
                else:
                    yield ev, elem
            elif ev is None:
                atrs = {
                    'type': typename,
                    'value': unicode(data['value'])
                }
                if 'name' in data:
                    atrs['name'] = data['name']
                if 'bin_offset' in data:
                    atrs['offset'] = unicode(data['bin_offset'])
                if 'bin_value' in data:
                    atrs['bin_value'] = unicode(data['bin_value'])
                if 'bin_type' in data:
                    atrs['type'] = data['bin_type'].__name__
                yield STARTEVENT, ('item', atrs)

                if isinstance(datatype, FlagsType):
                    fixed_size = datatype.basetype.fixed_size
                    b = bin(data['value'])[2:]
                    if len(b) < fixed_size * 8:
                        b = '0' * (fixed_size * 8 - len(b)) + b
                    h = hex(data['value'])[2:]
                    if len(h) < fixed_size * 2:
                        h = '0' * (fixed_size * 2 - len(h)) + h
                    if h.endswith('L'):
                        h = h[:-1]
                    atrs = {
                        'hex': h,
                        'bin': b
                    }
                    yield STARTEVENT, ('bitflags', atrs)

                    for bitfield_name in datatype.bitfields:
                        desc = datatype.bitfields[bitfield_name]
                        bitfield_type = desc.valuetype
                        value = desc.__get__(data['value'], None)

                        atrs = {
                            'type': datatype.basetype.__name__,
                            'name': bitfield_name,
                            'msb': unicode(desc.msb),
                            'lsb': unicode(desc.lsb),
                            'value': unicode(value)
                        }
                        yield STARTEVENT, ('bits', atrs)

                        if isinstance(bitfield_type, EnumType):
                            atrs = {
                                'type': bitfield_type.__name__,
                                'value': value.name
                            }
                            yield STARTEVENT, ('enum', atrs)
                            yield ENDEVENT, 'enum'

                        yield ENDEVENT, 'bits'

                    yield ENDEVENT, 'bitflags'

                yield ENDEVENT, 'item'



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals


HWPTAG_BEGIN = 0x010
tagnames = {
    # DocInfo Records
    HWPTAG_BEGIN + 0: 'HWPTAG_DOCUMENT_PROPERTIES',
    HWPTAG_BEGIN + 1: 'HWPTAG_ID_MAPPINGS',
    HWPTAG_BEGIN + 2: 'HWPTAG_BIN_DATA',
    HWPTAG_BEGIN + 3: 'HWPTAG_FACE_NAME',
    HWPTAG_BEGIN + 4: 'HWPTAG_BORDER_FILL',
    HWPTAG_BEGIN + 5: 'HWPTAG_CHAR_SHAPE',
    HWPTAG_BEGIN + 6: 'HWPTAG_TAB_DEF',
    HWPTAG_BEGIN + 7: 'HWPTAG_NUMBERING',
    HWPTAG_BEGIN + 8: 'HWPTAG_BULLET',
    HWPTAG_BEGIN + 9: 'HWPTAG_PARA_SHAPE',
    HWPTAG_BEGIN + 10: 'HWPTAG_STYLE',
    HWPTAG_BEGIN + 11: 'HWPTAG_DOC_DATA',
    HWPTAG_BEGIN + 12: 'HWPTAG_DISTRIBUTE_DOC_DATA',
    # HWPTAG_BEGIN + 13: RESERVED,
    HWPTAG_BEGIN + 14: 'HWPTAG_COMPATIBLE_DOCUMENT',
    HWPTAG_BEGIN + 15: 'HWPTAG_LAYOUT_COMPATIBILITY',
    HWPTAG_BEGIN + 16: 'HWPTAG_BEGIN_PLUS_16',

    # Section Records
    HWPTAG_BEGIN + 50: 'HWPTAG_PARA_HEADER',
    HWPTAG_BEGIN + 51: 'HWPTAG_PARA_TEXT',
    HWPTAG_BEGIN + 52: 'HWPTAG_PARA_CHAR_SHAPE',
    HWPTAG_BEGIN + 53: 'HWPTAG_PARA_LINE_SEG',
    HWPTAG_BEGIN + 54: 'HWPTAG_PARA_RANGE_TAG',
    HWPTAG_BEGIN + 55: 'HWPTAG_CTRL_HEADER',
    HWPTAG_BEGIN + 56: 'HWPTAG_LIST_HEADER',
    HWPTAG_BEGIN + 57: 'HWPTAG_PAGE_DEF',
    HWPTAG_BEGIN + 58: 'HWPTAG_FOOTNOTE_SHAPE',
    HWPTAG_BEGIN + 59: 'HWPTAG_PAGE_BORDER_FILL',
    HWPTAG_BEGIN + 60: 'HWPTAG_SHAPE_COMPONENT',
    HWPTAG_BEGIN + 61: 'HWPTAG_TABLE',
    HWPTAG_BEGIN + 62: 'HWPTAG_SHAPE_COMPONENT_LINE',
    HWPTAG_BEGIN + 63: 'HWPTAG_SHAPE_COMPONENT_RECTANGLE',
    HWPTAG_BEGIN + 64: 'HWPTAG_SHAPE_COMPONENT_ELLIPSE',
    HWPTAG_BEGIN + 65: 'HWPTAG_SHAPE_COMPONENT_ARC',
    HWPTAG_BEGIN + 66: 'HWPTAG_SHAPE_COMPONENT_POLYGON',
    HWPTAG_BEGIN + 67: 'HWPTAG_SHAPE_COMPONENT_CURVE',
    HWPTAG_BEGIN + 68: 'HWPTAG_SHAPE_COMPONENT_OLE',
    HWPTAG_BEGIN + 69: 'HWPTAG_SHAPE_COMPONENT_PICTURE',
    HWPTAG_BEGIN + 70: 'HWPTAG_SHAPE_COMPONENT_CONTAINER',
    HWPTAG_BEGIN + 71: 'HWPTAG_CTRL_DATA',
    HWPTAG_BEGIN + 72: 'HWPTAG_CTRL_EQEDIT',
    # HWPTAG_BEGIN + 73: RESERVED
    HWPTAG_BEGIN + 74: 'HWPTAG_SHAPE_COMPONENT_TEXTART',
    HWPTAG_BEGIN + 75: 'HWPTAG_FORM_OBJECT',
    HWPTAG_BEGIN + 76: 'HWPTAG_MEMO_SHAPE',
    HWPTAG_BEGIN + 77: 'HWPTAG_MEMO_LIST',
    HWPTAG_BEGIN + 78: 'HWPTAG_FORBIDDEN_CHAR',
    HWPTAG_BEGIN + 79: 'HWPTAG_CHART_DATA',
    # ...
    HWPTAG_BEGIN + 99: 'HWPTAG_SHAPE_COMPONENT_UNKNOWN',
}
for k, v in tagnames.items():
    globals()[v] = k
del k, v



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
import logging

from ..errors import InvalidOleStorageError


try:
    bytes
except NameError:
    bytes = str


logger = logging.getLogger(__name__)


def is_enabled():
    try:
        from gi.repository import Gsf
    except Exception:
        return False
    else:
        Gsf
        return True


def open(path):
    from gi.repository.Gsf import InputGio
    from gi.repository.Gsf import InfileMSOle

    inp = InputGio.new_for_path(path)
    return InfileMSOle.new(inp)


def listdir(gsfole):
    for i in xrange(gsfole.num_children()):
        yield gsfole.name_by_index(i)


class OleStorage:

    def __init__(self, gsfole):
        from gi.repository.Gsf import Input
        from gi.repository.Gsf import InfileMSOle

        if isinstance(gsfole, InfileMSOle):
            self.gsfole = gsfole
        elif isinstance(gsfole, Input):
            try:
                self.gsfole = InfileMSOle.new(gsfole)
            except Exception:
                raise InvalidOleStorageError()
        else:
            try:
                self.gsfole = open(gsfole)
            except Exception:
                raise InvalidOleStorageError()

    def __iter__(self):
        return listdir(self.gsfole)

    def __getitem__(self, name):
        child = self.gsfole.child_by_name(name)
        if child:
            if child.num_children() == -1:
                return OleStreamItem(self.gsfole, name)
            else:
                return OleStorage(child)
        else:
            raise KeyError(name)

    def close(self):
        del self.gsfole


class OleStreamItem:

    def __init__(self, parent, name):
        self.__parent = parent
        self.__name = name

    def open(self):
        gsfole = self.__parent.child_by_name(self.__name)
        if gsfole:
            return OleStream(gsfole)
        else:
            raise IOError(self.__name)


class OleStream:

    def __init__(self, gsfole):
        self.gsfole = gsfole

    def close(self):
        pass

    def read(self, size=None):
        pos = self.gsfole.tell()
        totalsize = self.gsfole.size
        if size is not None:
            if pos + size > totalsize:
                size = totalsize - pos
        else:
            size = totalsize - pos

        if size > 0:
            return self.gsfole.read(size)
        return bytes()

    def seek(self, offset, whence=0):
        from gi.repository.GLib import SeekType
        if whence == 0:
            whence = SeekType.SET
        elif whence == 1:
            whence = SeekType.CUR
        elif whence == 2:
            whence = SeekType.END
        else:
            raise ValueError(whence)

        self.gsfole.seek(offset, whence)

    def tell(self):
        return self.gsfole.tell()



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from six import with_metaclass

from hwp5.binmodel._shared import RecordModel
from hwp5.tagids import HWPTAG_PARA_LINE_SEG
from hwp5.binmodel._shared import ref_parent_member
from hwp5.dataio import ArrayType
from hwp5.dataio import Struct
from hwp5.dataio import UINT32
from hwp5.dataio import Flags
from hwp5.dataio import SHWPUNIT
from hwp5.dataio import INT32
from hwp5.dataio import X_ARRAY


class LineSeg(Struct):
    Flags = Flags(UINT32,
                  # 0, 'first_in_page',
                  # 1, 'first_in_column',
                  # 16, 'empty',
                  17, 'line_head',
                  18, 'line_tail',
                  # 19, 'auto_hyphen',
                  20, 'indented',
                  21, 'bullet')

    def attributes(cls):
        yield INT32, 'chpos',
        yield SHWPUNIT, 'y',
        yield SHWPUNIT, 'height',
        yield SHWPUNIT, 'height_text',
        yield SHWPUNIT, 'height_baseline',
        yield SHWPUNIT, 'space_below',
        yield SHWPUNIT, 'x',
        yield SHWPUNIT, 'width'
        yield cls.Flags, 'lineseg_flags'
    attributes = classmethod(attributes)


class ParaLineSeg(RecordModel):
    ''' 4.2.4. 문단의 레이아웃 '''

    tagid = HWPTAG_PARA_LINE_SEG

    def attributes(cls):
        ''' 표 57 문단의 레이아웃 '''
        yield dict(name='linesegs',
                   type=X_ARRAY(LineSeg, ref_parent_member('linesegs')))
    attributes = classmethod(attributes)


class ParaLineSegList(with_metaclass(ArrayType, list)):

    itemtype = LineSeg

    def read(cls, f, context):
        payload = context['stream'].read()
        return cls.decode(context, payload)
    read = classmethod(read)

    def decode(cls, context, payload):
        import struct
        unitfmt = 'iiiiiiiiHH'
        unitsize = struct.calcsize('<' + unitfmt)
        unitcount = len(payload) // unitsize
        values = struct.unpack('<' + unitfmt * unitcount, payload)
        names = ['chpos', 'y', 'height', 'height2', 'height85', 'space_below',
                 'x', 'width', 'a8', 'flags']
        x = list(dict(zip(names, tuple(values[i * 10:i * 10 + 10])))
                 for i in range(0, unitcount))
        for d in x:
            d['flags'] = LineSeg.Flags(d['flags'])
        return x
    decode = classmethod(decode)



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from hwp5.binmodel.controlchar import CHID
from hwp5.dataio import Enum
from hwp5.dataio import Flags
from hwp5.dataio import UINT32
from hwp5.dataio import WCHAR
from hwp5.binmodel.controls._shared import Control


class PageNumberPosition(Control):
    ''' 4.2.10.9. 쪽 번호 위치 '''
    chid = CHID.PGNP
    Position = Enum(NONE=0,
                    TOP_LEFT=1, TOP_CENTER=2, TOP_RIGHT=3,
                    BOTTOM_LEFT=4, BOTTOM_CENTER=5, BOTTOM_RIGHT=6,
                    OUTSIDE_TOP=7, OUTSIDE_BOTTOM=8,
                    INSIDE_TOP=9, INSIDE_BOTTOM=10)
    Flags = Flags(UINT32,
                  0, 7, 'shape',
                  8, 11, Position, 'position')

    def attributes(cls):
        yield cls.Flags, 'flags'
        yield WCHAR, 'usersymbol'
        yield WCHAR, 'prefix'
        yield WCHAR, 'suffix'
        yield WCHAR, 'dash'
    attributes = classmethod(attributes)



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from hwp5.binmodel._shared import RecordModel
from hwp5.tagids import HWPTAG_TAB_DEF
from hwp5.dataio import HWPUNIT
from hwp5.dataio import UINT32
from hwp5.dataio import Enum
from hwp5.dataio import Flags
from hwp5.dataio import Struct
from hwp5.dataio import N_ARRAY


class Tab(Struct):

    Kind = Enum(LEFT=0, RIGHT=1, CENTER=2, FLOAT=3)

    Flags = Flags(UINT32,
                  0, 7, Kind, 'kind',
                  8, 15, 'fill_type')

    @classmethod
    def attributes(cls):
        yield HWPUNIT, 'pos',
        yield cls.Flags, 'flags'


class TabDef(RecordModel):
    ''' 4.1.7. 탭 정의 '''

    tagid = HWPTAG_TAB_DEF

    ''' 표 32 탭 정의 속성 '''
    Flags = Flags(UINT32,
                  0, 'autotab_left',
                  1, 'autotab_right')

    @classmethod
    def attributes(cls):
        yield dict(type=cls.Flags, name='flags')
        yield dict(type=N_ARRAY(UINT32, Tab), name='tabs')



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
''' Extract out the specified stream in the <hwp5file> to the standard output.

Usage::

    hwp5proc cat [--loglevel=<loglevel>] [--logfile=<logfile>]
                 [--vstreams | --ole]
                 <hwp5file> <stream>
    hwp5proc cat --help

Options::

    -h --help               Show this screen
       --loglevel=<level>   Set log level.
       --logfile=<file>     Set log file.

       --vstreams           Process with virtual streams (i.e. parsed/converted
                            form of real streams)
       --ole                Treat <hwpfile> as an OLE Compound File. As a
                            result, some streams will be presented as-is. (i.e.
                            not decompressed)

Example::

    $ hwp5proc cat samples/sample-5017.hwp BinData/BIN0002.jpg | file -

    $ hwp5proc cat samples/sample-5017.hwp BinData/BIN0002.jpg > BIN0002.jpg

    $ hwp5proc cat samples/sample-5017.hwp PrvText | iconv -f utf-16le -t utf-8

    $ hwp5proc cat --vstreams samples/sample-5017.hwp PrvText.utf8

    $ hwp5proc cat --vstreams samples/sample-5017.hwp FileHeader.txt

    ccl: 0
    cert_drm: 0
    cert_encrypted: 0
    cert_signature_extra: 0
    cert_signed: 0
    compressed: 1
    distributable: 0
    drm: 0
    history: 0
    password: 0
    script: 0
    signature: HWP Document File
    version: 5.0.1.7
    xmltemplate_storage: 0

'''
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
import sys

from . import open_hwpfile
from ..storage import open_storage_item


PY2 = sys.version_info.major == 2


def main(args):
    if PY2:
        output_fp = sys.stdout
    else:
        output_fp = sys.stdout.buffer

    hwp5file = open_hwpfile(args)
    stream = open_storage_item(hwp5file, args['<stream>'])
    f = stream.open()
    try:
        while True:
            data = f.read(4096)
            if data:
                output_fp.write(data)
            else:
                return
    finally:
        if hasattr(f, 'close'):
            f.close()



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from hwp5.binmodel.controlchar import CHID
from hwp5.dataio import Enum
from hwp5.dataio import Flags
from hwp5.dataio import Struct
from hwp5.dataio import UINT16
from hwp5.dataio import WORD
from hwp5.dataio import HWPUNIT16
from hwp5.dataio import X_ARRAY
from hwp5.dataio import ref_member_flag
from hwp5.binmodel.controls._shared import Control
from hwp5.binmodel._shared import Border


class Column0(Struct):

    @staticmethod
    def attributes():
        yield WORD, 'width'


class Column(Struct):
    @staticmethod
    def attributes():
        yield WORD, 'spacing'
        yield WORD, 'width'


class ColumnsDef(Control):
    ''' 4.2.10.2. 단 정의 '''
    chid = CHID.COLD

    Kind = Enum('normal', 'distribute', 'parallel')
    Direction = Enum('l2r', 'r2l', 'both')
    Flags = Flags(UINT16,
                  0, 1, Kind, 'kind',
                  2, 9, 'count',
                  10, 11, Direction, 'direction',
                  12, 'same_widths')

    def attributes(cls):
        yield cls.Flags, 'flags'
        yield HWPUNIT16, 'spacing'

        def not_same_widths(context, values):
            ''' flags.same_widths == 0 '''
            return not values['flags'].same_widths

        def n_entries(member_ref):
            def n_entries(context, values):
                n_columns = member_ref(context, values)
                return n_columns - 1
            return n_entries

        yield dict(name='column0', type=Column0,
                   condition=not_same_widths)
        yield dict(name='columns',
                   type=X_ARRAY(Column,
                                n_entries(ref_member_flag('flags', 'count'))),
                   condition=not_same_widths)
        yield UINT16, 'attr2'
        yield Border, 'splitter'
    attributes = classmethod(attributes)



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from array import array
from binascii import b2a_hex
from itertools import takewhile
import inspect
import logging
import struct
import sys

from six import with_metaclass


PY3 = sys.version_info.major == 3


if PY3:
    long = int
    unicode = str
    basestring = str


logger = logging.getLogger(__name__)


class Eof(Exception):
    def __init__(self, *args):
        self.args = args


class OutOfData(Exception):
    pass


def readn(f, size):
    data = f.read(size)
    datasize = len(data)
    if datasize == 0:
        try:
            pos = f.tell()
        except IOError:
            pos = '<UNKNOWN>'
        raise Eof(pos)
    return data


class PrimitiveType(type):
    def __new__(mcs, name, bases, attrs):
        basetype = bases[0]
        attrs['basetype'] = basetype
        attrs.setdefault('__slots__', [])

        never_instantiate = attrs.pop('never_instantiate', True)
        if never_instantiate and '__new__' not in attrs:
            def __new__(cls, *args, **kwargs):
                return basetype.__new__(basetype, *args, **kwargs)
            attrs['__new__'] = __new__

        if 'binfmt' in attrs:
            binfmt = attrs['binfmt']
            fixed_size = struct.calcsize(binfmt)

            if 'fixed_size' in attrs:
                assert fixed_size == attrs['fixed_size']
            else:
                attrs['fixed_size'] = fixed_size

            if 'decode' not in attrs:
                def decode(cls, s):
                    return struct.unpack(binfmt, s)[0]
                attrs['decode'] = classmethod(decode)

        if 'fixed_size' in attrs and 'read' not in attrs:
            fixed_size = attrs['fixed_size']

            def read(cls, f):
                s = readn(f, fixed_size)
                decode = getattr(cls, 'decode', None)
                if decode:
                    return decode(s)
                return s
            attrs['read'] = classmethod(read)

        return type.__new__(mcs, str(name), bases, attrs)


def Primitive(name, basetype, binfmt, **attrs):
    attrs['binfmt'] = binfmt
    return PrimitiveType(name, (basetype,), attrs)


UINT32 = Primitive('UINT32', long, '<I')
INT32 = Primitive('INT32', int, '<i')
UINT16 = Primitive('UINT16', int, '<H')
INT16 = Primitive('INT16', int, '<h')
UINT8 = Primitive('UINT8', int, '<B')
INT8 = Primitive('INT8', int, '<b')
WORD = Primitive('WORD', int, '<H')
BYTE = Primitive('BYTE', int, '<B')
DOUBLE = Primitive('DOUBLE', float, '<d')
WCHAR = Primitive('WCHAR', int, '<H')
HWPUNIT = Primitive('HWPUNIT', long, '<I')
SHWPUNIT = Primitive('SHWPUNIT', int, '<i')
HWPUNIT16 = Primitive('HWPUNIT16', int, '<h')

inch2mm = lambda x: float(int(x * 25.4 * 100 + 0.5)) / 100  # noqa
hwp2inch = lambda x: x / 7200.0  # noqa
hwp2mm = lambda x: inch2mm(hwp2inch(x))  # noqa
hwp2pt = lambda x: int((x / 100.0) * 10 + 0.5) / 10.0  # noqa


class HexBytes(type):
    def __new__(mcs, size):
        decode = staticmethod(b2a_hex)
        return type.__new__(mcs, str('HexBytes(%d)' % size), (str,),
                            dict(fixed_size=size, decode=decode))


def decode_uint16le_array_default(bytes):
    codes = array(str('H'), bytes)
    if sys.byteorder == 'big':
        codes.byteswap()
    return codes


def decode_uint16le_array_in_jython(bytes):
    codes = array(str('h'), bytes)
    assert codes.itemsize == 2
    assert sys.byteorder == 'big'
    codes.byteswap()
    codes = array(str('H'), codes.tostring())
    assert codes.itemsize == 4
    return codes


in_jython = sys.platform.startswith('java')
if in_jython:
    decode_uint16le_array = decode_uint16le_array_in_jython
else:
    decode_uint16le_array = decode_uint16le_array_default


class BSTR(with_metaclass(PrimitiveType, unicode)):

    def read(f):
        size = UINT16.read(f)
        if size == 0:
            return u''
        data = readn(f, 2 * size)
        return decode_utf16le_with_hypua(data)
    read = staticmethod(read)


def decode_utf16le_with_hypua(bytes):
    ''' decode utf-16le encoded bytes with Hanyang-PUA codes into a unicode
    string with Hangul Jamo codes

    :param bytes: utf-16le encoded bytes with Hanyang-PUA codes
    :returns: a unicode string with Hangul Jamo codes
    '''
    return bytes.decode('utf-16le')


class BitGroupDescriptor(object):
    def __init__(self, bitgroup):
        valuetype = int
        if isinstance(bitgroup, tuple):
            if len(bitgroup) > 2:
                lsb, msb, valuetype = bitgroup
            else:
                lsb, msb = bitgroup
        else:
            lsb = msb = bitgroup
        self.lsb = lsb
        self.msb = msb
        self.valuetype = valuetype

    def __get__(self, instance, owner):
        valuetype = self.valuetype
        return valuetype(self.get_int_value(instance))

    def get_int_value(self, instance):
        lsb = self.lsb
        msb = self.msb
        return int(instance >> lsb) & int((2 ** (msb + 1 - lsb)) - 1)


class FlagsType(type):
    def __new__(mcs, name, bases, attrs):
        basetype = attrs.pop('basetype')
        bases = (basetype.basetype,)

        bitgroups = dict((k, BitGroupDescriptor(v))
                         for k, v in attrs.items())

        attrs = dict(bitgroups)
        attrs['__name__'] = name
        attrs['__slots__'] = ()

        attrs['basetype'] = basetype
        attrs['bitfields'] = bitgroups

        def dictvalue(self):
            return dict((name, getattr(self, name))
                        for name in bitgroups.keys())
        attrs['dictvalue'] = dictvalue

        return type.__new__(mcs, str(name), bases, attrs)


def _lex_flags_args(args):
    for idx, arg in enumerate(args):
        while True:
            pushback = (yield idx, arg)
            if pushback is arg:
                yield
                continue
            break


def _parse_flags_args(args):
    args = _lex_flags_args(args)
    try:
        idx = -1
        while True:
            # lsb
            try:
                idx, lsb = next(args)
            except StopIteration:
                break
            assert isinstance(lsb, int), ('#%d arg is expected to be'
                                          'a int: %s' % (idx, repr(lsb)))

            # msb (default: lsb)
            idx, x = next(args)
            if isinstance(x, int):
                msb = x
            elif isinstance(x, (type, basestring)):
                args.send(x)  # pushback
                msb = lsb
            else:
                assert False, '#%d arg is unexpected type: %s' % (idx, repr(x))

            # type (default: int)
            idx, x = next(args)
            assert not isinstance(x, int), ('#%d args is expected to be a type'
                                            'or name: %s' % (idx, repr(x)))
            if isinstance(x, type):
                t = x
            elif isinstance(x, basestring):
                args.send(x)  # pushback
                t = int
            else:
                assert False, '#%d arg is unexpected type: %s' % (idx, repr(x))

            # name
            idx, name = next(args)
            assert isinstance(name, basestring), ('#%d args is expected to be '
                                                  'a name: %s' % (idx,
                                                                  repr(name)))

            yield name, (lsb, msb, t)

    except StopIteration:
        assert False, '#%d arg is expected' % (idx + 1)


def Flags(basetype, *args):
    attrs = dict(_parse_flags_args(args))
    attrs['basetype'] = basetype
    return FlagsType('Flags', (), attrs)


enum_type_instances = set()


class EnumType(type):
    def __new__(mcs, enum_type_name, bases, attrs):
        items = attrs.pop('items')
        moreitems = attrs.pop('moreitems')

        populate_state = [1]

        names_by_instance = dict()
        instances_by_name = dict()
        instances_by_value = dict()

        def __new__(cls, value, name=None):
            if isinstance(value, cls):
                return value

            if name is None:
                if value in instances_by_value:
                    return instances_by_value[value]
                else:
                    logger.warning('undefined %s value: %s',
                                   cls.__name__, value)
                    logger.warning('defined name/values: %s',
                                   str(instances_by_name))
                    return int.__new__(cls, value)

            if len(populate_state) == 0:
                raise TypeError()

            assert name not in instances_by_name

            if value in instances_by_value:
                self = instances_by_value[value]
            else:
                # define new instance of this enum
                self = int.__new__(cls, value)
                instances_by_value[value] = self
                names_by_instance[self] = name

            instances_by_name[name] = self
            return self
        attrs['__new__'] = __new__
        attrs['__slots__'] = []
        attrs['scoping_struct'] = None

        class NameDescriptor(object):
            def __get__(self, instance, owner):
                if instance is None:
                    return owner.__name__
                return names_by_instance.get(instance)

        attrs['name'] = NameDescriptor()

        def __repr__(self):
            enum_name = type(self).__name__
            item_name = self.name
            if item_name is not None:
                return enum_name + '.' + item_name
            else:
                return '%s(%d)' % (enum_name, self)
        attrs['__repr__'] = __repr__

        cls = type.__new__(mcs, str(enum_type_name), bases, attrs)

        for v, k in enumerate(items):
            setattr(cls, k, cls(v, k))
        for k, v in moreitems.items():
            setattr(cls, k, cls(v, k))

        cls.names = set(instances_by_name.keys())
        cls.instances = set(names_by_instance.keys())

        # no more population
        populate_state.pop()

        enum_type_instances.add(cls)
        return cls

    def __init__(cls, *args, **kwargs):
        pass


def Enum(*items, **moreitems):
    attrs = dict(items=items, moreitems=moreitems)
    return EnumType('Enum', (int,), attrs)


class CompoundType(type):
    def __new__(mcs, name, bases, attrs):
        return type.__new__(mcs, str(name), bases, attrs)


class ArrayType(CompoundType):
    def __init__(self, *args, **kwargs):
        pass


class FixedArrayType(ArrayType):

    classes = dict()

    def __new__(mcs, itemtype, size):
        key = itemtype, size

        cls = mcs.classes.get(key)
        if cls is not None:
            return cls

        attrs = dict(itemtype=itemtype, size=size)
        name = 'ARRAY(%s,%s)' % (itemtype.__name__, size)
        cls = ArrayType.__new__(mcs, str(name), (tuple,), attrs)
        mcs.classes[key] = cls
        return cls


ARRAY = FixedArrayType


class VariableLengthArrayType(ArrayType):

    classes = dict()

    def __new__(mcs, counttype, itemtype):
        key = counttype, itemtype

        cls = mcs.classes.get(key)
        if cls is not None:
            return cls

        attrs = dict(itemtype=itemtype, counttype=counttype)
        name = 'N_ARRAY(%s,%s)' % (counttype.__name__, itemtype.__name__)
        cls = ArrayType.__new__(mcs, str(name), (list,), attrs)
        mcs.classes[key] = cls
        return cls


N_ARRAY = VariableLengthArrayType


def ref_member(member_name):
    def fn(context, values):
        return values[member_name]
    fn.__doc__ = member_name
    return fn


def ref_member_flag(member_name, bitfield_name):
    def fn(context, values):
        return getattr(values[member_name], bitfield_name)
    fn.__doc__ = '%s.%s' % (member_name, bitfield_name)
    return fn


class X_ARRAY(object):

    def __init__(self, itemtype, count_reference):
        name = 'ARRAY(%s, \'%s\')' % (itemtype.__name__,
                                      count_reference.__doc__)
        self.__doc__ = self.__name__ = name
        self.itemtype = itemtype
        self.count_reference = count_reference

    def __call__(self, context, values):
        count = self.count_reference(context, values)
        return ARRAY(self.itemtype, count)


class SelectiveType(object):

    def __init__(self, selector_reference, selections):
        self.__name__ = 'SelectiveType'
        self.selections = selections
        self.selector_reference = selector_reference

    def __call__(self, context, values):
        selector = self.selector_reference(context, values)
        return self.selections.get(selector, Struct)  # default: empty struct


class ParseError(Exception):

    treegroup = None

    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)
        self.cause = None
        self.path = None
        self.record = None
        self.binevents = None
        self.parse_stack_traces = []

    def print_to_logger(self, logger):
        e = self
        logger.error('ParseError: %s', e)
        logger.error('Caused by: %s', repr(e.cause))
        logger.error('Path: %s', e.path)
        if e.treegroup is not None:
            logger.error('Treegroup: %s', e.treegroup)
        if e.record:
            logger.error('Record: %s', e.record['seqno'])
            logger.error('Record Payload:')
            for line in dumpbytes(e.record['payload'], True):
                logger.error('  %s', line)
        logger.error('Problem Offset: at %d (=0x%x)', e.offset, e.offset)
        if self.binevents:
            logger.error('Binary Parse Events:')
            from hwp5.bintype import log_events
            for ev, item in log_events(self.binevents, logger.error):
                pass
        logger.error('Model Stack:')
        for level, c in enumerate(reversed(e.parse_stack_traces)):
            model = c['model']
            if isinstance(model, StructType):
                logger.error('  %s', model)
                parsed_members = c['parsed']
                for member in parsed_members:
                    offset = member.get('offset', 0)
                    offset_end = member.get('offset_end', 1)
                    name = member['name']
                    value = member['value']
                    logger.error('    %06x:%06x: %s = %s',
                                 offset, offset_end - 1, name, value)
                logger.error('    %06x:      : %s', c['offset'], c['member'])
                pass
            else:
                logger.error('  %s%s', ' ' * level, c)


def typed_struct_attributes(struct, attributes, context):
    attributes = dict(attributes)

    def popvalue(member):
        name = member['name']
        if name in attributes:
            return attributes.pop(name)
        else:
            return member['type']()

    for member in struct.parse_members_with_inherited(context, popvalue):
        yield member

    # remnants
    for name, value in attributes.items():
        yield dict(name=name, type=type(value), value=value)


class StructType(CompoundType):
    def __init__(cls, name, bases, attrs):
        super(StructType, cls).__init__(name, bases, attrs)
        if 'attributes' in cls.__dict__:
            members = (dict(type=member[0], name=member[1])
                       if isinstance(member, tuple)
                       else member
                       for member in cls.attributes())
            cls.members = list(members)
        for k, v in attrs.items():
            if isinstance(v, EnumType):
                v.__name__ = k
                v.scoping_struct = cls
            elif isinstance(v, FlagsType):
                v.__name__ = k

    def parse_members(cls, context, getvalue):
        if 'attributes' not in cls.__dict__:
            return
        values = dict()
        for member in cls.members:
            member = dict(member)
            if isinstance(member['type'], X_ARRAY):
                member['type'] = member['type'](context, values)
            elif isinstance(member['type'], SelectiveType):
                member['type'] = member['type'](context, values)

            member_version = member.get('version')
            if member_version is None or context['version'] >= member_version:
                condition_func = member.get('condition')
                if condition_func is None or condition_func(context, values):
                    try:
                        value = getvalue(member)
                    except ParseError as e:
                        tracepoint = dict(model=cls, member=member['name'])
                        e.parse_stack_traces.append(tracepoint)
                        raise
                    values[member['name']] = member['value'] = value
                    yield member

    def parse_members_with_inherited(cls, context, getvalue, up_to_cls=None):
        mro = inspect.getmro(cls)
        mro = takewhile(lambda cls: cls is not up_to_cls, mro)
        mro = list(cls for cls in mro if 'attributes' in cls.__dict__)
        mro = reversed(mro)
        for cls in mro:
            for member in cls.parse_members(context, getvalue):
                yield member


class Struct(with_metaclass(StructType, object)):
    pass


def dumpbytes(data, crust=False):
    if PY3:
        _ord = int
    else:
        _ord = ord

    offsbase = 0
    if crust:
        yield '\t 0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F'
    while len(data) > 16:
        if crust:
            line = '%05x0: ' % offsbase
        else:
            line = ''
        line += ' '.join(['%02x' % _ord(ch) for ch in data[0:16]])
        yield line
        data = data[16:]
        offsbase += 1

    if crust:
        line = '%05x0: ' % offsbase
    else:
        line = ''
    line += ' '.join(['%02x' % _ord(ch) for ch in data])
    yield line


def hexdump(data, crust=False):
    return '\n'.join([line for line in dumpbytes(data, crust)])


class IndentedOutput:
    def __init__(self, base, level):
        self.base = base
        self.level = level

    def write(self, x):
        for line in x.split('\n'):
            if len(line) > 0:
                self.base.write('\t' * self.level)
                self.base.write(line)
                self.base.write('\n')


class Printer:
    def __init__(self, baseout):
        self.baseout = baseout

    def prints(self, *args):
        for x in args:
            self.baseout.write(str(x) + ' ')
        self.baseout.write('\n')



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from hwp5.binmodel._shared import RecordModel
from hwp5.tagids import HWPTAG_STYLE
from hwp5.dataio import Enum
from hwp5.dataio import Flags
from hwp5.dataio import BYTE
from hwp5.dataio import BSTR
from hwp5.dataio import INT16
from hwp5.dataio import UINT16


class Style(RecordModel):
    ''' 4.1.11. 스타일 '''
    tagid = HWPTAG_STYLE

    # 표 43 스타일 종류
    Kind = Enum(PARAGRAPH=0, CHAR=1)
    Flags = Flags(BYTE,
                  0, 1, Kind, 'kind')

    def attributes(cls):
        ''' 표 42 스타일 '''
        # yield BSTR, 'local_name',
        yield BSTR, 'name',
        yield BSTR, 'local_name',
        yield cls.Flags, 'flags',
        yield BYTE, 'next_style_id',
        yield INT16, 'lang_id',
        yield UINT16, 'parashape_id',
        yield UINT16, 'charshape_id',

        # unknown fields
        # following fields are found from 5.0.0.0
        yield UINT16, 'unknown'
    attributes = classmethod(attributes)



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2017 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from contextlib import closing
import io
import logging
import os.path
import sys


logger = logging.getLogger(__name__)


def is_enabled():
    if not sys.platform.startswith('java'):
        logger.info('%s: disabled', __name__)
        return False
    try:
        import javax.xml.transform
        javax
    except ImportError:
        logger.info('%s: disabled', __name__)
        return False
    else:
        logger.info('%s: enabled', __name__)
        return True


def xslt(xsl_path, inp_path, out_path):
    transform = xslt_compile(xsl_path)
    with io.open(out_path, 'wb') as f:
        return transform(inp_path, f)


class XSLT:

    def __init__(self, xsl_path, **params):
        from javax.xml.transform import URIResolver
        from javax.xml.transform import TransformerFactory
        from javax.xml.transform.stream import StreamSource
        from java.io import FileInputStream

        xsl_path = os.path.abspath(xsl_path)
        xsl_base = os.path.dirname(xsl_path)

        xsl_fis = FileInputStream(xsl_path)

        xsl_source = StreamSource(xsl_fis)

        class BaseURIResolver(URIResolver):

            def __init__(self, base):
                self.base = base

            def resolve(self, href, base):
                path = os.path.join(self.base, href)
                path = os.path.abspath(path)
                fis = FileInputStream(path)
                return StreamSource(fis)

        uri_resolver = BaseURIResolver(xsl_base)

        xslt_factory = TransformerFactory.newInstance()
        xslt_factory.setURIResolver(uri_resolver)

        self.transformer = xslt_factory.newTransformer(xsl_source)
        for k, v in params.items():
            self.transformer.setParameter(k, unicode(v))

    def transform(self, input, output):
        '''
        >>> T.transform('input.xml', 'output.xml')
        '''
        from java.io import FileInputStream
        from java.io import FileOutputStream
        out_path = os.path.abspath(output)
        inp_path = os.path.abspath(input)
        with closing(FileInputStream(inp_path)) as inp_fis:
            with closing(FileOutputStream(out_path)) as out_fos:
                return self._transform(inp_fis, out_fos)

    def transform_into_stream(self, input, output):
        '''
        >>> T.transform('input.xml', sys.stdout)
        '''
        from java.io import FileInputStream
        inp_path = os.path.abspath(input)
        with closing(FileInputStream(inp_path)) as inp_fis:
            out_fos = wrap_filelike_outputstream(output)
            return self._transform(inp_fis, out_fos)

    def _transform(self, input, output):
        from javax.xml.transform.stream import StreamSource
        from javax.xml.transform.stream import StreamResult
        inp_source = StreamSource(input)
        out_result = StreamResult(output)
        self.transformer.transform(inp_source, out_result)
        return dict()


def xslt_compile(xsl_path, **params):
    xslt = XSLT(xsl_path, **params)
    return xslt.transform_into_stream


def wrap_filelike_inputstream(f):
    from org.python.core import FilelikeInputStream
    return FilelikeInputStream(f)


def wrap_filelike_outputstream(f):
    from java.io import OutputStream

    class FilelikeOutputStream(OutputStream):

        def write(self, *args):
            if len(args) == 1:
                # byte
                ch = chr(args[0] & 0xff)
                f.write(ch)
            if len(args) == 3:
                # array.array, offset, length
                array, offset, length = args
                buf = array.tostring()
                f.write(buf[offset:offset+length])
            else:
                logger.debug('%r', args)
                self.super__write(*args)

        def flush(self):
            f.flush()

        def close(self):
            pass
    return FilelikeOutputStream()



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import logging

from six import with_metaclass

from hwp5.dataio import StructType
from hwp5.dataio import Struct
from hwp5.dataio import Enum
from hwp5.dataio import Flags
from hwp5.dataio import PrimitiveType
from hwp5.dataio import UINT32
from hwp5.dataio import UINT16
from hwp5.dataio import UINT8
from hwp5.dataio import INT32
from hwp5.dataio import INT8
from hwp5.dataio import BYTE
from hwp5.dataio import HWPUNIT16
from hwp5.dataio import SHWPUNIT
from hwp5.dataio import N_ARRAY


logger = logging.getLogger(__name__)


tag_models = dict()


class RecordModelType(StructType):

    def __new__(mcs, name, bases, attrs):
        cls = StructType.__new__(mcs, name, bases, attrs)
        if 'tagid' in attrs:
            tagid = attrs['tagid']
            assert tagid not in tag_models
            tag_models[tagid] = cls
        return cls


class RecordModel(with_metaclass(RecordModelType, object)):
    pass


class BinStorageId(UINT16):
    pass


class COLORREF(with_metaclass(PrimitiveType, int)):
    binfmt = INT32.binfmt
    never_instantiate = False

    def __getattr__(self, name):
        if name == 'r':
            return self & 0xff
        elif name == 'g':
            return (self & 0xff00) >> 8
        elif name == 'b':
            return (self & 0xff0000) >> 16
        elif name == 'a':
            return int((self & 0xff000000) >> 24)
        elif name == 'rgb':
            return self.r, self.g, self.b

    def __str__(self):
        return '#%02x%02x%02x' % (self.r, self.g, self.b)

    def __repr__(self):
        class_name = self.__class__.__name__
        value = '(0x%02x, 0x%02x, 0x%02x)' % self.rgb
        return class_name + value


class Margin(Struct):
    def attributes():
        yield HWPUNIT16, 'left'
        yield HWPUNIT16, 'right'
        yield HWPUNIT16, 'top'
        yield HWPUNIT16, 'bottom'
    attributes = staticmethod(attributes)


class Coord(Struct):
    def attributes():
        yield SHWPUNIT, 'x'
        yield SHWPUNIT, 'y'
    attributes = staticmethod(attributes)


class Border(Struct):

    # 표 20 테두리선 종류
    StrokeEnum = Enum('none', 'solid',
                      'dashed', 'dotted', 'dash-dot', 'dash-dot-dot',
                      'long-dash', 'large-dot',
                      'double', 'double-2', 'double-3', 'triple',
                      'wave', 'double-wave',
                      'inset', 'outset', 'groove', 'ridge')
    StrokeType = Flags(UINT8,
                       0, 4, StrokeEnum, 'stroke_type')

    # 표 21 테두리선 굵기
    widths = {'0.1mm': 0,
              '0.12mm': 1,
              '0.15mm': 2,
              '0.2mm': 3,
              '0.25mm': 4,
              '0.3mm': 5,
              '0.4mm': 6,
              '0.5mm': 7,
              '0.6mm': 8,
              '0.7mm': 9,
              '1.0mm': 10,
              '1.5mm': 11,
              '2.0mm': 12,
              '3.0mm': 13,
              '4.0mm': 14,
              '5.0mm': 15}
    WidthEnum = Enum(**widths)
    Width = Flags(UINT8,
                  0, 4, WidthEnum, 'width')

    def attributes(cls):
        yield cls.StrokeType, 'stroke_flags',
        yield cls.Width, 'width_flags',
        yield COLORREF, 'color',
    attributes = classmethod(attributes)


class BorderLine(Struct):
    ''' 표 81 테두리 선 정보 '''

    LineEnd = Enum('round', 'flat')
    ArrowShape = Enum('none', 'arrow', 'arrow2', 'diamond', 'circle', 'rect',
                      'diamondfilled', 'disc', 'rectfilled')
    ArrowSize = Enum('smallest', 'smaller', 'small', 'abitsmall', 'normal',
                     'abitlarge', 'large', 'larger', 'largest')

    ''' 표 82 테두리 선 정보 속성 '''
    Flags = Flags(UINT32,
                  0, 5, Border.StrokeEnum, 'stroke',
                  6, 9, LineEnd, 'line_end',
                  10, 15, ArrowShape, 'arrow_start',
                  16, 21, ArrowShape, 'arrow_end',
                  22, 25, ArrowSize, 'arrow_start_size',
                  26, 29, ArrowSize, 'arrow_end_size',
                  30, 'arrow_start_fill',
                  31, 'arrow_end_fill')

    def attributes(cls):
        yield COLORREF, 'color'
        yield INT32, 'width'
        yield cls.Flags, 'flags'
    attributes = classmethod(attributes)


class Fill(Struct):
    pass


class FillNone(Fill):
    def attributes():
        yield UINT32, 'size',  # SPEC is confusing
    attributes = staticmethod(attributes)


class FillColorPattern(Fill):
    ''' 표 23 채우기 정보 '''
    PatternTypeEnum = Enum(NONE=255, HORIZONTAL=0, VERTICAL=1, BACKSLASH=2,
                           SLASH=3, GRID=4, CROSS=5)
    PatternTypeFlags = Flags(UINT32,
                             0, 7, PatternTypeEnum, 'pattern_type')

    def attributes(cls):
        yield COLORREF, 'background_color',
        yield COLORREF, 'pattern_color',
        yield cls.PatternTypeFlags, 'pattern_type_flags',
    attributes = classmethod(attributes)


class FillImage(Fill):
    FillImageEnum = Enum(TILE_ALL=0,
                         TILE_HORIZONTAL_TOP=1,
                         TILE_HORIZONTAL_BOTTOM=2,
                         TILE_VERTICAL_LEFT=3,
                         TILE_VERTICAL_RIGHT=4,
                         RESIZE=5,
                         CENTER=6,
                         CENTER_TOP=7,
                         CENTER_BOTTOM=8,
                         LEFT_MIDDLE=9,
                         LEFT_TOP=10,
                         LEFT_BOTTOM=11,
                         RIGHT_MIDDLE=12,
                         RIGHT_TOP=13,
                         RIGHT_BOTTOM=14,
                         NONE=15)
    FillImageFlags = Flags(BYTE,
                           0, 16, FillImageEnum, 'fillimage_type')

    EffectEnum = Enum(REAL_PIC=0,
                      GRAY_SCALE=1,
                      BLACK_WHITE=2,
                      PATTERN8x8=3)
    EffectFlags = Flags(UINT8,
                        0, 8, EffectEnum, 'effect_type')

    def attributes(cls):
        yield cls.FillImageFlags, 'flags'
        yield INT8, 'brightness'
        yield INT8, 'contrast'
        yield cls.EffectFlags, 'effect'
        yield UINT16, 'bindata_id'
    attributes = classmethod(attributes)


class Coord32(Struct):
    def attributes():
        yield UINT32, 'x'
        yield UINT32, 'y'
    attributes = staticmethod(attributes)


GradationTypeEnum = Enum(LINEAR=1, CIRCULAR=2, CONICAL=3, RECTANGULAR=4)
GradationTypeFlags = Flags(BYTE,
                           0, 8, GradationTypeEnum, 'gradation_type')


class FillGradation(Fill):
    def attributes():
        yield GradationTypeFlags, 'type',
        yield UINT32, 'shear',  # 기울임 각 (도)
        yield Coord32, 'center',
        yield UINT32, 'blur',  # 번짐 정도: 0-100
        # TODO: 스펙 1.2에 따르면 색상 수 > 2인 경우
        # 색상이 바뀌는 위치 배열이 온다고 함
        yield N_ARRAY(UINT32, COLORREF), 'colors',
    attributes = staticmethod(attributes)


def ref_parent_member(member_name):
    def f(context, values):
        context, model = context['parent']
        return model['content'][member_name]
    f.__doc__ = 'PARENTREC.' + member_name
    return f



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
'''Do various operations on HWPv5 files.

Usage::

    hwp5proc <command> [<args>...]
    hwp5proc [--version]
    hwp5proc [--help]
    hwp5proc [--help-commands]

       --version        Show version and copyright information.
    -h --help           Show help messages.
       --help-commands  Show available commands.
'''
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
import gettext
import logging
import os
import sys

from docopt import docopt

from .. import __version__
from ..dataio import ParseError
from ..errors import InvalidHwp5FileError
from ..plat import xsltproc
from ..plat import xmllint
from ..storage import ExtraItemStorage
from ..storage import open_storage_item
from ..storage.ole import OleStorage
from ..xmlmodel import Hwp5File


PY3 = sys.version_info.major == 3
logger = logging.getLogger(__name__)

locale_dir = os.path.join(os.path.dirname(__file__), '..', 'locale')
locale_dir = os.path.abspath(locale_dir)
t = gettext.translation('hwp5proc', locale_dir, fallback=True)
if PY3:
    _ = t.gettext
else:
    _ = t.ugettext


def rest_to_docopt(doc):
    ''' ReST to docopt conversion
    '''
    return doc.replace('::\n\n', ':\n').replace('``', '')


def init_with_environ():
    if 'PYHWP_XSLTPROC' in os.environ:
        xsltproc.executable = os.environ['PYHWP_XSLTPROC']
        xsltproc.enable()

    if 'PYHWP_XMLLINT' in os.environ:
        xmllint.executable = os.environ['PYHWP_XMLLINT']
        xmllint.enable()


def init_logger(args):
    logger = logging.getLogger('hwp5')
    try:
        from colorlog import ColoredFormatter
    except ImportError:
        formatter = None
    else:
        formatter = ColoredFormatter(
            '%(log_color)s%(levelname)-8s%(reset)s %(blue)s%(message)s',
            datefmt=None, reset=True,
            log_colors={
                'DEBUG': 'cyan',
                'INFO': 'green',
                'WARNING': 'yellow',
                'ERROR': 'red',
                'CRITICAL': 'red'
            }
        )

    loglevel = args.get('--loglevel', None)
    if not loglevel:
        loglevel = os.environ.get('PYHWP_LOGLEVEL')
    if loglevel:
        levels = dict(debug=logging.DEBUG,
                      info=logging.INFO,
                      warning=logging.WARNING,
                      error=logging.ERROR,
                      critical=logging.CRITICAL)
        loglevel = loglevel.lower()
        loglevel = levels.get(loglevel, logging.WARNING)
        logger.setLevel(loglevel)

    logfile = args.get('--logfile', None)
    if not logfile:
        logfile = os.environ.get('PYHWP_LOGFILE')
    if logfile:
        handler = logging.FileHandler(logfile)
    else:
        handler = logging.StreamHandler()
    if formatter:
        handler.setFormatter(formatter)
    logger.addHandler(handler)


subcommands = [
    'version',
    'header',
    'summaryinfo',
    'ls',
    'cat',
    'unpack',
    'records',
    'models',
    'find',
    'xml',
    'rawunz',
    'diststream',
]


program = 'hwp5proc (pyhwp) {version}'.format(version=__version__)

copyright = 'Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>'

license = _('''License AGPLv3+: GNU Affero GPL version 3 or any later
<http://gnu.org/licenses/agpl.txt>.
This is free software: you are free to change and redistribute it.
There is NO WARRANTY, to the extent permitted by law.''')

disclosure = _('''Disclosure: This program has been developed in accordance with a public
document named "HWP Binary Specification 1.1" published by Hancom Inc.
<http://www.hancom.co.kr>.''')  # noqa

version = '''{program}
{copyright}
{license}
{disclosure}'''.format(
    program=program,
    copyright=copyright,
    license=license,
    disclosure=disclosure,
)


def print_subcommands():
    print(_('Available <command> values:'))
    print('')
    for subcommand in subcommands:
        print('    {}'.format(subcommand))
    print('')
    print(_('See \'hwp5proc <command> --help\' '
            'for more information on a specific command.'))


def main():
    doc = __doc__
    doc = rest_to_docopt(doc)

    args = docopt(doc, version=version, help=False, options_first=True)

    if args['--help-commands']:
        print(doc)
        print_subcommands()
        return 0

    command = args['<command>']
    if command is None:
        print(doc.strip())
        return 1

    if command not in subcommands:
        message = _('Unknown command: {}')
        message = message.format(command)
        message = '\n{}\n\n'.format(message)
        sys.stderr.write(message)
        print_subcommands()
        return 1

    argv = [command] + args['<args>']
    mod = __import__('hwp5.proc.' + command, fromlist=['main'])
    main = mod.main
    doc = rest_to_docopt(mod.__doc__)
    args = docopt(doc, version=__version__, argv=argv)
    init_logger(args)

    try:
        return main(args)
    except InvalidHwp5FileError as e:
        logger.error('%s', e)
        raise SystemExit(1)
    except ParseError as e:
        e.print_to_logger(logger)
        raise SystemExit(1)


def open_hwpfile(args):
    filename = args['<hwp5file>']
    if args['--ole']:
        hwpfile = OleStorage(filename)
    else:
        hwpfile = Hwp5File(filename)
        if args['--vstreams']:
            hwpfile = ExtraItemStorage(hwpfile)
    return hwpfile


def parse_recordstream_name(hwpfile, streamname):
    if streamname == 'docinfo':
        return hwpfile.docinfo
    segments = streamname.split('/')
    if len(segments) == 2:
        if segments[0] == 'bodytext':
            try:
                idx = int(segments[1])
                return hwpfile.bodytext.section(idx)
            except ValueError:
                pass
    return open_storage_item(hwpfile, streamname)



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from io import BytesIO
import logging
import sys

from .bintype import read_type
from .compressed import decompress
from .dataio import UINT32, Flags, Struct
from .errors import InvalidOleStorageError
from .errors import InvalidHwp5FileError
from .storage import ItemWrapper
from .storage import StorageWrapper
from .storage import ItemConversionStorage
from .storage import is_stream
from .storage.ole import OleStorage
from .summaryinfo import CLSID_HWP_SUMMARY_INFORMATION
from .utils import GeneratorTextReader
from .utils import cached_property
from .utils import transcoder

PY3 = sys.version_info.major == 3
if PY3:
    basestring = str


logger = logging.getLogger(__name__)


HWP5_SIGNATURE = b'HWP Document File' + (b'\x00' * 15)


class BYTES(type):
    def __new__(mcs, size):
        decode = staticmethod(lambda bytes, *args, **kwargs: bytes)
        return type.__new__(mcs, str('BYTES(%d)') % size, (str,),
                            dict(fixed_size=size, decode=decode))


class VERSION(object):
    fixed_size = 4

    if PY3:
        def decode(cls, bytes):
            return (bytes[3], bytes[2], bytes[1], bytes[0])
    else:
        def decode(cls, bytes):
            return (ord(bytes[3]), ord(bytes[2]),
                    ord(bytes[1]), ord(bytes[0]))
    decode = classmethod(decode)


class FileHeader(Struct):
    Flags = Flags(UINT32,
                  0, 'compressed',
                  1, 'password',
                  2, 'distributable',
                  3, 'script',
                  4, 'drm',
                  5, 'xmltemplate_storage',
                  6, 'history',
                  7, 'cert_signed',
                  8, 'cert_encrypted',
                  9, 'cert_signature_extra',
                  10, 'cert_drm',
                  11, 'ccl')

    def attributes(cls):
        yield BYTES(32), 'signature'
        yield VERSION, 'version'
        yield cls.Flags, 'flags'
        yield BYTES(216), 'reserved'
    attributes = classmethod(attributes)


def is_hwp5file(filename):
    ''' Test whether it is an HWP format v5 file. '''
    try:
        olestg = OleStorage(filename)
    except InvalidOleStorageError:
        return False
    return storage_is_hwp5file(olestg)


def storage_is_hwp5file(stg):
    try:
        fileheader = stg['FileHeader']
    except KeyError:
        logger.info('stg has no FileHeader')
        return False
    fileheader = HwpFileHeader(fileheader)
    if fileheader.signature == HWP5_SIGNATURE:
        return True
    else:
        logger.info('fileheader.signature = %r', fileheader.signature)
        return False


class CompressedStream(ItemWrapper):

    def open(self):
        return decompress(self.wrapped.open())


class CompressedStorage(StorageWrapper):
    ''' decompress streams in the underlying storage '''
    def __getitem__(self, name):
        item = self.wrapped[name]
        if is_stream(item):
            return CompressedStream(item)
        else:
            return item


class PasswordProtectedStream(ItemWrapper):

    def open(self):
        # TODO: 현재로선 암호화된 내용을 그냥 반환
        logger.warning('Password-encrypted stream: currently decryption is '
                       'not supported')
        return self.wrapped.open()


class PasswordProtectedStorage(StorageWrapper):
    def __getitem__(self, name):
        item = self.wrapped[name]
        if is_stream(item):
            return PasswordProtectedStream(item)
        else:
            return item


class Hwp5PasswordProtectedDoc(ItemConversionStorage):

    def resolve_conversion_for(self, name):
        if name in ('BinData', 'BodyText', 'Scripts', 'ViewText'):
            return PasswordProtectedStorage
        elif name in ('DocInfo', ):
            return PasswordProtectedStream


class VersionSensitiveItem(ItemWrapper):

    def __init__(self, item, version):
        ItemWrapper.__init__(self, item)
        self.version = version

    def open(self):
        return self.wrapped.open()

    def other_formats(self):
        return dict()


class Hwp5FileBase(ItemConversionStorage):
    ''' Base of an Hwp5File.

    Hwp5FileBase checks basic validity of an HWP format v5 and provides
    `fileheader` property.

    :param stg: an OLE2 structured storage.
    :type stg: an instance of storage, OleFileIO or filename
    :raises InvalidHwp5FileError: `stg` is not a valid HWP format v5 document.
    '''

    def __init__(self, stg):
        if isinstance(stg, basestring):
            try:
                stg = OleStorage(stg)
            except InvalidOleStorageError:
                raise InvalidHwp5FileError('Not an OLE2 Compound Binary File.')

        if not storage_is_hwp5file(stg):
            errormsg = 'Not an HWP Document format v5 storage.'
            raise InvalidHwp5FileError(errormsg)

        ItemConversionStorage.__init__(self, stg)

    def resolve_conversion_for(self, name):
        if name == 'FileHeader':
            return HwpFileHeader

    def get_fileheader(self):
        return self['FileHeader']

    fileheader = cached_property(get_fileheader)

    header = fileheader


class Hwp5DistDocStream(VersionSensitiveItem):

    def open(self):
        from hwp5.distdoc import decode
        encodedstream = self.wrapped.open()
        return decode(encodedstream)

    def head_record(self):
        item = self.wrapped.open()
        from .recordstream import read_record
        return read_record(item, 0)

    def head_record_stream(self):
        from .recordstream import record_to_json
        record = self.head_record()
        json = record_to_json(record)
        return GeneratorTextReader(iter([json]))

    def head(self):
        record = self.head_record()
        return record['payload']

    def head_stream(self):
        return BytesIO(self.head())

    def head_sha1(self):
        from hwp5.distdoc import decode_head_to_sha1
        payload = self.head()
        return decode_head_to_sha1(payload)

    def head_key(self):
        from hwp5.distdoc import decode_head_to_key
        payload = self.head()
        return decode_head_to_key(payload)

    def tail(self):
        item = self.wrapped.open()
        from .recordstream import read_record
        read_record(item, 0)
        assert 4 + 256 == item.tell()
        return item.read()

    def tail_decrypted(self):
        from hwp5.distdoc import decrypt_tail
        key = self.head_key()
        tail = self.tail()
        return decrypt_tail(key, tail)

    def tail_stream(self):
        return BytesIO(self.tail())


class Hwp5DistDocStorage(ItemConversionStorage):

    def resolve_conversion_for(self, name):
        def conversion(item):
            return Hwp5DistDocStream(self.wrapped[name], None)  # TODO: version
        return conversion


class Hwp5DistDoc(ItemConversionStorage):

    def resolve_conversion_for(self, name):
        if name in ('Scripts', 'ViewText'):
            return Hwp5DistDocStorage


class Hwp5Compression(ItemConversionStorage):
    ''' handle compressed streams in HWPv5 files '''

    def resolve_conversion_for(self, name):
        if name in ('BinData', 'BodyText', 'ViewText'):
            return CompressedStorage
        elif name == 'DocInfo':
            return CompressedStream
        elif name == 'Scripts':
            return CompressedStorage


class PreviewText(object):

    def __init__(self, item):
        self.open = item.open

    def other_formats(self):
        return {'.utf8': self.open_utf8}

    def open_utf8(self):
        transcode = transcoder('utf-16le', 'utf-8')
        return transcode(self.open())

    def get_utf8(self):
        f = self.open_utf8()
        try:
            return f.read()
        finally:
            f.close()

    utf8 = cached_property(get_utf8)

    def get_text(self):
        fp = self.open()
        try:
            data = fp.read()
        finally:
            fp.close()
        return data.decode('utf-16le')

    text = cached_property(get_text)

    def __str__(self):
        if PY3:
            return self.text
        return self.utf8

    def __unicode__(self):
        return self.text


class Sections(ItemConversionStorage):

    section_class = VersionSensitiveItem

    def __init__(self, stg, version):
        ItemConversionStorage.__init__(self, stg)
        self.version = version

    def resolve_conversion_for(self, name):
        def conversion(item):
            return self.section_class(self.wrapped[name], self.version)
        return conversion

    def other_formats(self):
        return dict()

    def section(self, idx):
        return self['Section%d' % idx]

    def section_indexes(self):
        def gen():
            for name in self:
                if name.startswith('Section'):
                    idx = name[len('Section'):]
                    try:
                        idx = int(idx)
                    except:
                        pass
                    else:
                        yield idx
        indexes = list(gen())
        indexes.sort()
        return indexes

    @property
    def sections(self):
        return list(self.section(idx)
                    for idx in self.section_indexes())


class HwpFileHeader(object):

    def __init__(self, item):
        self.open = item.open

    def to_dict(self):
        f = self.open()
        try:
            return read_type(FileHeader, dict(), f)
        finally:
            f.close()

    value = cached_property(to_dict)

    def get_version(self):
        return self.value['version']

    version = cached_property(get_version)

    def get_signature(self):
        return self.value['signature']

    signature = cached_property(get_signature)

    def get_flags(self):
        return FileHeader.Flags(self.value['flags'])

    flags = cached_property(get_flags)

    def open_text(self):
        signature = self.value['signature']
        signature = signature.decode('latin1')
        signature = signature[:len('HWP Document File')]

        d = FileHeader.Flags.dictvalue(self.value['flags'])
        d['signature'] = signature
        d['version'] = '%d.%d.%d.%d' % self.value['version']
        out = BytesIO()
        for k, v in sorted(d.items()):
            out.write('{}: {}\n'.format(k, v).encode('utf-8'))
        out.seek(0)
        return out

    def other_formats(self):
        return {'.txt': self.open_text}


class HwpSummaryInfo(VersionSensitiveItem):

    def other_formats(self):
        return {'.txt': self.open_text}

    def getPropertySetStream(self):
        from .msoleprops import PropertySetFormat
        from .msoleprops import PropertySetStreamReader
        from .summaryinfo import FMTID_HWP_SUMMARY_INFORMATION
        from .summaryinfo import HWP_PROPERTIES

        propertySetFormat = PropertySetFormat(
            FMTID_HWP_SUMMARY_INFORMATION,
            HWP_PROPERTIES
        )
        reader = PropertySetStreamReader([propertySetFormat])
        f = self.open()
        try:
            return reader.read(f)
        finally:
            f.close()

    propertySetStream = cached_property(getPropertySetStream)

    def getHwpSummaryInfoPropertySet(self):
        stream = self.propertySetStream
        if stream.clsid == CLSID_HWP_SUMMARY_INFORMATION:
            return stream.propertysets[0]

    propertySet = cached_property(getHwpSummaryInfoPropertySet)

    @property
    def title(self):
        from .msoleprops import PIDSI_TITLE
        return self.propertySet[PIDSI_TITLE]

    @property
    def subject(self):
        from .msoleprops import PIDSI_SUBJECT
        return self.propertySet[PIDSI_SUBJECT]

    @property
    def author(self):
        from .msoleprops import PIDSI_AUTHOR
        return self.propertySet[PIDSI_AUTHOR]

    @property
    def keywords(self):
        from .msoleprops import PIDSI_KEYWORDS
        return self.propertySet[PIDSI_KEYWORDS]

    @property
    def comments(self):
        from .msoleprops import PIDSI_COMMENTS
        return self.propertySet[PIDSI_COMMENTS]

    @property
    def lastSavedBy(self):
        from .msoleprops import PIDSI_LASTAUTHOR
        return self.propertySet[PIDSI_LASTAUTHOR]

    @property
    def revisionNumber(self):
        from .msoleprops import PIDSI_REVNUMBER
        return self.propertySet[PIDSI_REVNUMBER]

    @property
    def lastPrintedTime(self):
        from .msoleprops import PIDSI_LASTPRINTED
        return self.propertySet[PIDSI_LASTPRINTED]

    @property
    def createdTime(self):
        from .msoleprops import PIDSI_CREATE_DTM
        return self.propertySet[PIDSI_CREATE_DTM]

    @property
    def lastSavedTime(self):
        from .msoleprops import PIDSI_LASTSAVE_DTM
        return self.propertySet[PIDSI_LASTSAVE_DTM]

    @property
    def numberOfPages(self):
        from .msoleprops import PIDSI_PAGECOUNT
        return self.propertySet[PIDSI_PAGECOUNT]

    @property
    def dateString(self):
        from .summaryinfo import HWPPIDSI_DATE_STR
        return self.propertySet[HWPPIDSI_DATE_STR]

    @property
    def numberOfParagraphs(self):
        from .summaryinfo import HWPPIDSI_PARACOUNT
        return self.propertySet[HWPPIDSI_PARACOUNT]

    @property
    def plaintext_lines(self):
        from .msoleprops import PropertySetStreamTextFormatter
        stream = self.getPropertySetStream()
        formatter = PropertySetStreamTextFormatter()
        return formatter.formatTextLines(stream)

    def open_text(self):
        out = BytesIO()
        for line in self.plaintext_lines:
            line = line.encode('utf-8')
            out.write(line + b'\n')
        out.seek(0)
        return out


class Hwp5File(ItemConversionStorage):
    ''' represents HWPv5 File

        Hwp5File(stg)

        stg: an instance of Storage
    '''

    def __init__(self, stg):
        stg = Hwp5FileBase(stg)

        if stg.header.flags.password:
            stg = Hwp5PasswordProtectedDoc(stg)

            # TODO: 현재로선 decryption이 구현되지 않았으므로,
            # 레코드 파싱은 불가능하다. 적어도 encrypted stream에
            # 직접 접근은 가능하도록, 다음 레이어들은 bypass한다.
            ItemConversionStorage.__init__(self, stg)
            return

        if stg.header.flags.distributable:
            stg = Hwp5DistDoc(stg)

        if stg.header.flags.compressed:
            stg = Hwp5Compression(stg)

        ItemConversionStorage.__init__(self, stg)

    def resolve_conversion_for(self, name):
        if name == 'DocInfo':
            return self.with_version(self.docinfo_class)
        if name == 'BodyText':
            return self.with_version(self.bodytext_class)
        if name == 'ViewText':
            return self.with_version(self.bodytext_class)
        if name == 'PrvText':
            return PreviewText
        if name == '\005HwpSummaryInformation':
            return self.with_version(self.summaryinfo_class)

    def with_version(self, f):
        def wrapped(item):
            return f(item, self.header.version)
        return wrapped

    summaryinfo_class = HwpSummaryInfo
    docinfo_class = VersionSensitiveItem
    bodytext_class = Sections

    @cached_property
    def summaryinfo(self):
        return self['\005HwpSummaryInformation']

    @cached_property
    def docinfo(self):
        return self['DocInfo']

    @cached_property
    def preview_text(self):
        return self['PrvText']

    @cached_property
    def bodytext(self):
        return self['BodyText']

    @cached_property
    def viewtext(self):
        return self['ViewText']

    @property
    def text(self):
        if self.header.flags.distributable:
            return self.viewtext
        else:
            return self.bodytext



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from hwp5.binmodel._shared import RecordModel
from hwp5.tagids import HWPTAG_DOCUMENT_PROPERTIES
from hwp5.dataio import UINT16
from hwp5.dataio import UINT32


class DocumentProperties(RecordModel):
    ''' 4.1.1. 문서 속성 '''

    tagid = HWPTAG_DOCUMENT_PROPERTIES

    def attributes():
        ''' 표 9 문서 속성 '''
        yield UINT16, 'section_count',
        yield UINT16, 'page_startnum',
        yield UINT16, 'footnote_startnum',
        yield UINT16, 'endnote_startnum',
        yield UINT16, 'picture_startnum',
        yield UINT16, 'table_startnum',
        yield UINT16, 'math_startnum',
        yield UINT32, 'list_id',
        yield UINT32, 'paragraph_id',
        yield UINT32, 'character_unit_loc_in_paragraph',
        # yield UINT32, 'flags',   # DIFFSPEC
    attributes = staticmethod(attributes)



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
import io
import os.path
import sys


PY3 = sys.version_info.major == 3
if PY3:
    basestring = str


def is_storage(item):
    return hasattr(item, '__iter__') and hasattr(item, '__getitem__')


def is_stream(item):
    return hasattr(item, 'open') and callable(item.open)


class ItemWrapper(object):
    def __init__(self, wrapped):
        self.wrapped = wrapped

    def __getattr__(self, name):
        return getattr(self.wrapped, name)


class StorageWrapper(ItemWrapper):
    def __iter__(self):
        return iter(self.wrapped)

    def __getitem__(self, name):
        return self.wrapped[name]


class ItemConversionStorage(StorageWrapper):

    def __getitem__(self, name):
        item = self.wrapped[name]
        # 기반 스토리지에서 찾은 아이템에 대해, conversion()한다.
        conversion = self.resolve_conversion_for(name)
        if conversion:
            return conversion(item)
        return item

    def resolve_conversion_for(self, name):
        ''' return a conversion function for the specified storage item '''
        pass


class ExtraItemStorage(StorageWrapper):

    def __iter__(self):
        for name in self.wrapped:
            yield name

            item = self.wrapped[name]
            if hasattr(item, 'other_formats'):
                other_formats = item.other_formats()
                if other_formats:
                    for ext in other_formats:
                        yield name + ext

    def __getitem__(self, name):
        try:
            item = self.wrapped[name]
            if is_storage(item):
                item = ExtraItemStorage(item)
            return item
        except KeyError:
            # 기반 스토리지에는 없으므로, other_formats() 중에서 찾아본다.
            for root in self.wrapped:
                item = self.wrapped[root]
                if hasattr(item, 'other_formats'):
                    other_formats = item.other_formats()
                    if other_formats:
                        for ext, func in other_formats.items():
                            if root + ext == name:
                                return Open2Stream(func)
            raise


class Open2Stream(object):

    def __init__(self, open):
        self.open = open


def iter_storage_leafs(stg, basepath=''):
    ''' iterate every leaf nodes in the storage

        stg: an instance of Storage
    '''
    for name in stg:
        path = basepath + name
        item = stg[name]
        if is_storage(item):
            for x in iter_storage_leafs(item, path + '/'):
                yield x
        else:
            yield path


def unpack(stg, outbase):
    ''' unpack a storage into outbase directory

        stg: an instance of Storage
        outbase: path to a directory in filesystem (should not end with '/')
    '''
    for name in stg:
        outpath = os.path.join(outbase, name)
        item = stg[name]
        if is_storage(item):
            if not os.path.exists(outpath):
                os.mkdir(outpath)
            unpack(item, outpath)
        else:
            f = item.open()
            try:
                outpath = outpath.replace('\x05', '_05')
                with io.open(outpath, 'wb') as outfile:
                    outfile.write(f.read())
            finally:
                f.close()


def open_storage_item(stg, path):
    if isinstance(path, basestring):
        path_segments = path.split('/')
    else:
        path_segments = path

    item = stg
    for name in path_segments:
        item = item[name]
    return item


def printstorage(stg, basepath=''):
    names = list(stg)
    names.sort()
    for name in names:
        path = basepath + name
        item = stg[name]
        if is_storage(item):
            printstorage(item, path + '/')
        elif is_stream(item):
            print(path.encode('unicode_escape').decode('utf-8'))



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from six import with_metaclass

from hwp5.binmodel._shared import RecordModelType
from hwp5.binmodel._shared import RecordModel
from hwp5.tagids import HWPTAG_LIST_HEADER
from hwp5.dataio import Enum
from hwp5.dataio import Flags
from hwp5.dataio import UINT32
from hwp5.dataio import UINT16
from hwp5.dataio import SHWPUNIT
from hwp5.dataio import HWPUNIT
from hwp5.dataio import HWPUNIT16
from hwp5.dataio import BYTE
from hwp5.binmodel._shared import Margin
from hwp5.binmodel.controls.table_control import TableControl
from hwp5.binmodel.controls.gshape_object_control import GShapeObjectControl
from hwp5.binmodel.controls.header_footer import Header
from hwp5.binmodel.controls.header_footer import Footer
from hwp5.binmodel.tagid60_shape_component import ShapeComponent


list_header_models = dict()


class ListHeaderType(RecordModelType):

    def __new__(mcs, name, bases, attrs):
        cls = RecordModelType.__new__(mcs, name, bases, attrs)
        if 'parent_model_type' in attrs:
            parent_model_type = attrs['parent_model_type']
            before_tablebody = attrs.get('before_tablebody', False)
            list_type_key = parent_model_type, before_tablebody
            assert list_type_key not in list_header_models
            list_header_models[list_type_key] = cls
        return cls


class ListHeader(with_metaclass(ListHeaderType, RecordModel)):
    ''' 4.2.7. 문단 리스트 헤더 '''

    tagid = HWPTAG_LIST_HEADER

    VAlign = Enum(TOP=0, MIDDLE=1, BOTTOM=2)
    Flags = Flags(UINT32,
                  0, 2, 'textdirection',
                  3, 4, 'linebreak',
                  5, 6, VAlign, 'valign')

    def attributes(cls):
        ''' 표 60 문단 리스트 헤더 '''
        yield UINT16, 'paragraphs',
        yield UINT16, 'unknown1',
        yield cls.Flags, 'listflags',
    attributes = classmethod(attributes)

    extension_types = list_header_models

    def get_extension_key(context, model):
        ''' (parent model type, after TableBody) '''
        if 'parent' in context:
            context, model = context['parent']
            seen_table_body = context.get('seen_table_body', False)
            return model['type'], seen_table_body
    get_extension_key = staticmethod(get_extension_key)


class TableCaption(ListHeader):
    ''' 표 66 캡션 리스트 '''
    parent_model_type = TableControl
    before_tablebody = False

    # 표 68 캡션 속성
    Position = Enum(LEFT=0, RIGHT=1, TOP=2, BOTTOM=3)
    Flags = Flags(UINT32,
                  0, 1, Position, 'position',
                  # 캡션이 긴 경우 테이블 여백 영역까지 확장
                  2, 'expand_to_margin')

    def attributes(cls):
        ''' 표 67 캡션 '''
        yield cls.Flags, 'flags',
        yield HWPUNIT, 'width',
        yield HWPUNIT16, 'separation',  # 캡션과 틀 사이 간격
        yield HWPUNIT, 'max_width',  # expand_to_margin 고려한 최대 너비
    attributes = classmethod(attributes)


class TableCell(ListHeader):
    ''' 표 75 셀 속성 '''
    parent_model_type = TableControl
    before_tablebody = True

    def attributes():
        yield UINT16, 'col',
        yield UINT16, 'row',
        yield UINT16, 'colspan',
        yield UINT16, 'rowspan',
        yield SHWPUNIT, 'width',
        yield SHWPUNIT, 'height',
        yield Margin, 'padding',
        yield UINT16, 'borderfill_id',
        yield SHWPUNIT, 'unknown_width',
    attributes = staticmethod(attributes)


class GShapeObjectCaption(TableCaption):
    parent_model_type = GShapeObjectControl


class TextboxParagraphList(ListHeader):
    ''' 표 85 그리기 개체 글상자용 텍스트 속성 '''
    parent_model_type = ShapeComponent

    def attributes():
        yield Margin, 'padding'
        yield HWPUNIT, 'maxwidth'
    attributes = staticmethod(attributes)


class HeaderFooterParagraphList(ListHeader):
    ''' 표 129 머리말/꼬리말 '''
    def attributes():
        yield HWPUNIT, 'width'
        yield HWPUNIT, 'height'
        yield BYTE, 'textrefsbitmap'
        yield BYTE, 'numberrefsbitmap'
    attributes = staticmethod(attributes)


class HeaderParagraphList(HeaderFooterParagraphList):
    parent_model_type = Header


class FooterParagraphList(HeaderFooterParagraphList):
    parent_model_type = Footer



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from hwp5.binmodel._shared import RecordModel
from hwp5.tagids import HWPTAG_TABLE
from hwp5.dataio import Enum
from hwp5.dataio import Flags
from hwp5.dataio import UINT32
from hwp5.dataio import UINT16
from hwp5.dataio import HWPUNIT16
from hwp5.dataio import N_ARRAY
from hwp5.dataio import X_ARRAY
from hwp5.dataio import ref_member
from hwp5.dataio import Struct
from hwp5.binmodel._shared import Margin


class ZoneInfo(Struct):
    def attributes():
        ''' 표 73 영역 속성 '''
        yield UINT16, 'starting_column'
        yield UINT16, 'starting_row'
        yield UINT16, 'end_column'
        yield UINT16, 'end_row'
        yield UINT16, 'borderfill_id'
    attributes = staticmethod(attributes)


class TableBody(RecordModel):
    ''' 4.2.9.1. 표 개체 '''
    tagid = HWPTAG_TABLE

    # 표 71 표 속성의 속성
    Split = Enum(NONE=0, BY_CELL=1, SPLIT=2)
    Flags = Flags(UINT32,
                  0, 1, Split, 'split_page',
                  2, 'repeat_header')

    def attributes(cls):
        ''' 표 70 표 개체 속성 '''
        yield cls.Flags, 'flags'
        yield UINT16, 'rows'
        yield UINT16, 'cols'
        yield HWPUNIT16, 'cellspacing'

        # 표 72 안쪽 여백 정보
        yield Margin, 'padding'

        yield dict(type=X_ARRAY(UINT16, ref_member('rows')),
                   name='rowcols')
        yield UINT16, 'borderfill_id'
        yield dict(type=N_ARRAY(UINT16, ZoneInfo),
                   name='validZones',
                   version=(5, 0, 0, 7))
    attributes = classmethod(attributes)



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
''' Decode distribute docs.

Based on the algorithm described by Changwoo Ryu
See https://groups.google.com/forum/#!topic/hwp-foss/d2KL2ypR89Q
'''
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from io import BytesIO
import logging

from .plat import get_aes128ecb_decrypt
from .recordstream import read_record
from .tagids import HWPTAG_DISTRIBUTE_DOC_DATA

logger = logging.getLogger(__name__)


def decode(stream):
    distdoc_data_record = read_record(stream, 0)
    if distdoc_data_record['tagid'] != HWPTAG_DISTRIBUTE_DOC_DATA:
        raise IOError('the first record is not an HWPTAG_DISTRIBUTE_DOC_DATA')
    distdoc_data = distdoc_data_record['payload']
    key = decode_head_to_key(distdoc_data)
    tail = stream.read()
    decrypted = decrypt_tail(key, tail)
    return BytesIO(decrypted)


class Random:
    ''' MSVC's srand()/rand() like pseudorandom generator.
    '''

    def __init__(self, seed):
        self.seed = seed

    def rand(self):
        self.seed = (self.seed * 214013 + 2531011) & 0xffffffff
        value = (self.seed >> 16) & 0x7fff
        return value


def decode_head_to_sha1(record_payload):
    ''' Decode HWPTAG_DISTRIBUTE_DOC_DATA.

    It's the sha1 digest of user-supplied password string, i.e.,

        '12345' -> hashlib.sha1('12345').digest()

    '''
    if len(record_payload) != 256:
        raise ValueError('payload size must be 256 bytes')

    data = bytearray(record_payload)
    seed = data[3] << 24 | data[2] << 16 | data[1] << 8 | data[0]
    random = Random(seed)

    n = 0
    for i in range(256):
        if n == 0:
            key = random.rand() & 0xff
            n = (random.rand() & 0xf) + 1
        if i >= 4:
            data[i] = data[i] ^ key
        n -= 1

    # decoded = b''.join(chr(x) for x in data)
    decoded = data
    sha1offset = 4 + (seed & 0xf)

    ucs16le = decoded[sha1offset:sha1offset + 80]
    return ucs16le


def decode_head_to_key(record_payload):
    sha1ucs16le = decode_head_to_sha1(record_payload)
    return sha1ucs16le[:16]


def decrypt_tail(key, encrypted_tail):
    decrypt = get_aes128ecb_decrypt()
    return decrypt(key, encrypted_tail)



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
'''HWPv5 to text converter

Usage::

    hwp5txt [options] <hwp5file>
    hwp5txt -h | --help
    hwp5txt --version

Options::

    -h --help           Show this screen
    --version           Show version
    --loglevel=<level>  Set log level.
    --logfile=<file>    Set log file.

    --output=<file>     Output file
'''
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from contextlib import closing
import gettext
import logging
import os.path
import sys

from docopt import docopt

from . import __version__ as version
from .dataio import ParseError
from .errors import InvalidHwp5FileError
from .proc import init_logger
from .proc import rest_to_docopt
from .utils import make_open_dest_file
from .utils import cached_property
from .transforms import BaseTransform
from .xmlmodel import Hwp5File


PY3 = sys.version_info.major == 3
logger = logging.getLogger(__name__)
locale_dir = os.path.join(os.path.dirname(__file__), 'locale')
locale_dir = os.path.abspath(locale_dir)
t = gettext.translation('hwp5txt', locale_dir, fallback=True)
if PY3:
    _ = t.gettext
else:
    _ = t.ugettext


RESOURCE_PATH_XSL_TEXT = 'xsl/plaintext.xsl'


class TextTransform(BaseTransform):

    @property
    def transform_hwp5_to_text(self):
        transform_xhwp5 = self.transform_xhwp5_to_text
        return self.make_transform_hwp5(transform_xhwp5)

    @cached_property
    def transform_xhwp5_to_text(self):
        '''
        >>> T.transform_xhwp5_to_css('hwp5.xml', 'styles.css')
        '''
        resource_path = RESOURCE_PATH_XSL_TEXT
        return self.make_xsl_transform(resource_path)


def main():

    doc = rest_to_docopt(__doc__)
    args = docopt(doc, version=version)
    init_logger(args)

    hwp5path = args['<hwp5file>']

    text_transform = TextTransform()

    open_dest = make_open_dest_file(args['--output'])
    transform = text_transform.transform_hwp5_to_text

    try:
        with closing(Hwp5File(hwp5path)) as hwp5file:
            with open_dest() as dest:
                transform(hwp5file, dest)
    except ParseError as e:
        e.print_to_logger(logger)
    except InvalidHwp5FileError as e:
        logger.error('%s', e)
        sys.exit(1)



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from hwp5.binmodel.controls._shared import Control
from hwp5.dataio import Enum
from hwp5.dataio import Flags
from hwp5.dataio import UINT32
from hwp5.dataio import SHWPUNIT
from hwp5.dataio import HWPUNIT
from hwp5.dataio import INT16
from hwp5.dataio import BSTR
from hwp5.binmodel._shared import Margin


class CommonControl(Control):

    # 표 65 개체 공통 속성의 속성
    Flow = Enum(FLOAT=0, BLOCK=1, BACK=2, FRONT=3)
    TextSide = Enum(BOTH=0, LEFT=1, RIGHT=2, LARGER=3)
    VRelTo = Enum(PAPER=0, PAGE=1, PARAGRAPH=2)
    HRelTo = Enum(PAPER=0, PAGE=1, COLUMN=2, PARAGRAPH=3)
    VAlign = Enum(TOP=0, MIDDLE=1, BOTTOM=2)
    HAlign = Enum(LEFT=0, CENTER=1, RIGHT=2, INSIDE=3, OUTSIDE=4)
    WidthRelTo = Enum(PAPER=0, PAGE=1, COLUMN=2, PARAGRAPH=3, ABSOLUTE=4)
    HeightRelTo = Enum(PAPER=0, PAGE=1, ABSOLUTE=2)
    NumberCategory = Enum(NONE=0, FIGURE=1, TABLE=2, EQUATION=3)

    CommonControlFlags = Flags(UINT32,
                               0, 'inline',
                               2, 'affect_line_spacing',
                               3, 4, VRelTo, 'vrelto',
                               5, 7, VAlign, 'valign',
                               8, 9, HRelTo, 'hrelto',
                               10, 12, HAlign, 'halign',
                               13, 'restrict_in_page',
                               14, 'overlap_others',
                               15, 17, WidthRelTo, 'width_relto',
                               18, 19, HeightRelTo, 'height_relto',
                               20, 'protect_size_when_vrelto_paragraph',
                               21, 23, Flow, 'flow',
                               24, 25, TextSide, 'text_side',
                               26, 27, NumberCategory, 'number_category')

    MARGIN_LEFT = 0
    MARGIN_RIGHT = 1
    MARGIN_TOP = 2
    MARGIN_BOTTOM = 3

    def attributes(cls):
        ''' 표 64 개체 공통 속성 '''
        yield cls.CommonControlFlags, 'flags',
        yield SHWPUNIT, 'y',    # DIFFSPEC
        yield SHWPUNIT, 'x',    # DIFFSPEC
        yield HWPUNIT, 'width',
        yield HWPUNIT, 'height',
        yield INT16, 'z_order',
        yield INT16, 'unknown1',
        yield Margin, 'margin',
        yield UINT32, 'instance_id',
        yield dict(type=INT16, name='unknown2', version=(5, 0, 0, 5))
        yield dict(type=BSTR, name='description', version=(5, 0, 0, 5))
    attributes = classmethod(attributes)



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
import io
import logging
import os.path
import sys

from ...errors import InvalidOleStorageError


PY3 = sys.version_info.major == 3
if PY3:
    basestring = str


logger = logging.getLogger(__name__)

enabled = False


def is_enabled():
    if 'PYHWP_PLAT_UNO' in os.environ:
        PYHWP_PLAT_UNO = os.environ['PYHWP_PLAT_UNO'].strip()
        try:
            forced = bool(int(PYHWP_PLAT_UNO))
            logger.debug('%s: forced to be %s by PYHWP_PLAT_UNO', __name__,
                         'enabled' if forced else 'disabled')
            return forced
        except:
            logger.warning('PYHWP_PLAT_UNO=%s (invalid)', PYHWP_PLAT_UNO)
    logger.debug('%s: is %s', __name__, 'enabled' if enabled else 'disabled')
    return enabled


def enable():
    import uno
    import unohelper

    g = globals()
    g['uno'] = uno
    g['unohelper'] = unohelper
    g['enabled'] = True
    logger.info('%s: enabled.', __name__)


def disable():
    global enabled
    enabled = False
    logger.info('%s: disabled.', __name__)


def XSLTTransformer(context, stylesheet_url, source_url, source_url_base):
    from com.sun.star.beans import NamedValue
    from hwp5.plat._uno.services import css
    args = (NamedValue('StylesheetURL', stylesheet_url),
            NamedValue('SourceURL', source_url),
            NamedValue('SourceBaseURL', source_url_base))
    select = os.environ.get('PYHWP_PLAT_UNO_XSLT', 'libxslt')
    logger.debug('PYHWP_PLAT_UNO_XSLT = %s', select)
    if select == 'jaxthelper':
        logger.debug('%s.xslt: using css.comp.JAXTHelper', __name__)
        return css.comp.JAXTHelper(context, *args)
    else:
        logger.debug('%s.xslt: using %s', __name__,
                     'css.comp.documentconversion.LibXSLTTransformer')
        return css.comp.documentconversion.LibXSLTTransformer(context, *args)


class OneshotEvent(object):

    def __init__(self):
        pin, pout = os.pipe()
        self.pin = os.fdopen(pin, 'r')
        self.pout = os.fdopen(pout, 'w')

    def wait(self):
        self.pin.read()
        self.pin.close()

    def signal(self):
        self.pout.close()


class XSLT(object):

    def __init__(self, context):
        self.context = context

    def __call__(self, xsl_path, inp_path, out_path):
        import uno
        import unohelper
        from hwp5.plat._uno import ucb
        from hwp5.plat._uno.adapters import OutputStreamToFileLike
        xsl_path = os.path.abspath(xsl_path)
        xsl_url = uno.systemPathToFileUrl(xsl_path)

        inp_path = os.path.abspath(inp_path)
        inp_url = uno.systemPathToFileUrl(inp_path)
        inp_stream = ucb.open_url(self.context, inp_url)

        out_path = os.path.abspath(out_path)
        with io.open(out_path, 'wb') as out_file:
            out_stream = OutputStreamToFileLike(out_file, dontclose=True)

            from com.sun.star.io import XStreamListener

            class XSLTListener(unohelper.Base, XStreamListener):
                def __init__(self):
                    self.event = OneshotEvent()

                def started(self):
                    logger.info('XSLT started')

                def closed(self):
                    logger.info('XSLT closed')
                    self.event.signal()

                def terminated(self):
                    logger.info('XSLT terminated')
                    self.event.signal()

                def error(self, exception):
                    logger.error('XSLT error: %s', exception)
                    self.event.signal()

                def disposing(self, source):
                    logger.info('XSLT disposing: %s', source)
                    self.event.signal()

            listener = XSLTListener()
            transformer = XSLTTransformer(self.context, xsl_url, '', '')
            transformer.InputStream = inp_stream
            transformer.OutputStream = out_stream
            transformer.addListener(listener)

            transformer.start()
            xsl_name = os.path.basename(xsl_path)
            logger.info('xslt.soffice(%s) start', xsl_name)
            try:
                listener.event.wait()
            finally:
                logger.info('xslt.soffice(%s) end', xsl_name)

            transformer.removeListener(listener)
            return dict()


def xslt(xsl_path, inp_path, out_path):
    import uno
    context = uno.getComponentContext()
    xslt = XSLT(context)
    return xslt(xsl_path, inp_path, out_path)


def oless_from_filename(filename):
    inputstream = inputstream_from_filename(filename)
    return oless_from_inputstream(inputstream)


def inputstream_from_filename(filename):
    f = io.open(filename, 'rb')
    from hwp5.plat._uno.adapters import InputStreamFromFileLike
    return InputStreamFromFileLike(f)


def oless_from_inputstream(inputstream):
    import uno
    context = uno.getComponentContext()
    sm = context.ServiceManager
    name = 'com.sun.star.embed.OLESimpleStorage'
    args = (inputstream, )
    return sm.createInstanceWithArgumentsAndContext(name, args, context)


class OleStorage(object):

    def __init__(self, stg):
        ''' an OLESimpleStorage to hwp5 storage adapter.

        :param stg: a filename or an instance of OLESimpleStorage
        '''
        if isinstance(stg, basestring):
            self.oless = oless_from_filename(stg)
            try:
                self.oless.getElementNames()
            except:
                errormsg = 'Not a valid OLE2 Compound Binary File.'
                raise InvalidOleStorageError(errormsg)
        else:
            # TODO assert stg is an instance of OLESimpleStorage
            self.oless = stg

    def __iter__(self):
        return iter(self.oless.getElementNames())

    def __getitem__(self, name):
        from com.sun.star.container import NoSuchElementException
        try:
            elem = self.oless.getByName(name)
        except NoSuchElementException:
            raise KeyError(name)
        services = elem.SupportedServiceNames
        if 'com.sun.star.embed.OLESimpleStorage' in services:
            return OleStorage(elem)
        else:
            elem.closeInput()
            return OleStorageStream(self.oless, name)

    def close(self):
        return
        # TODO
        # if this is root, close underlying olefile
        if self.path == '':
            # old version of OleFileIO has no close()
            if hasattr(self.olefile, 'close'):
                self.olefile.close()


class OleStorageStream(object):

    def __init__(self, oless, name):
        self.oless = oless
        self.name = name

    def open(self):
        from hwp5.plat._uno.adapters import FileFromStream
        stream = self.oless.getByName(self.name)
        return FileFromStream(stream)



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from hwp5.binmodel._shared import RecordModel
from hwp5.tagids import HWPTAG_SHAPE_COMPONENT_PICTURE
from hwp5.dataio import Struct
from hwp5.dataio import Flags
from hwp5.dataio import SHWPUNIT
from hwp5.dataio import UINT32
from hwp5.dataio import UINT16
from hwp5.dataio import INT8
from hwp5.dataio import BYTE
from hwp5.binmodel._shared import Margin
from hwp5.binmodel._shared import Coord
from hwp5.binmodel._shared import BorderLine


class PictureInfo(Struct):
    ''' 표 27 그림 정보 '''
    def attributes():
        yield INT8, 'brightness',
        yield INT8, 'contrast',
        yield BYTE, 'effect',
        yield UINT16, 'bindata_id',
    attributes = staticmethod(attributes)


class PictureEffect(Struct):
    ''' 표 103 그림 효과 속성 '''

    Flags = Flags(UINT32)

    @classmethod
    def attributes(cls):
        yield cls.Flags, 'flags'
        # TODO


# HWPML에서의 이름 사용
class ImageRect(Struct):
    ''' 이미지 좌표 정보 '''

    def attributes():
        yield Coord, 'p0'
        yield Coord, 'p1'
        yield Coord, 'p2'
        yield Coord, 'p3'
    attributes = staticmethod(attributes)


# HWPML에서의 이름 사용
class ImageClip(Struct):
    ''' 이미지 자르기 정보 '''

    def attributes():
        yield SHWPUNIT, 'left',
        yield SHWPUNIT, 'top',
        yield SHWPUNIT, 'right',
        yield SHWPUNIT, 'bottom',
    attributes = staticmethod(attributes)


class ShapePicture(RecordModel):
    ''' 4.2.9.4. 그림 개체 '''
    tagid = HWPTAG_SHAPE_COMPONENT_PICTURE

    def attributes():
        ''' 표 102 그림 개체 속성 '''
        yield BorderLine, 'border'
        yield ImageRect, 'rect',
        yield ImageClip, 'clip',
        yield Margin, 'padding',
        yield PictureInfo, 'picture',
        yield dict(type=BYTE, name='border_transparency', version=(5, 0, 2, 2))
        yield dict(type=UINT32, name='instance_id', version=(5, 0, 2, 5))

        # TODO: this choke on 5.0.3.3 d6dfac424525298119de54410c3b22d74aa85511
        # Strangely, its ok on 5.0.3.3 83a0ea1f9da368ff9f0b45f72e9306b776edf38a
        # and other 5.0.3.0, 5.0.3.2 and 5.0.3.4 files.
        yield dict(type=PictureEffect, name='picture_effect',
                   version=(5, 0, 3, 4))

    attributes = staticmethod(attributes)
